@@ -22,14 +22,14 @@
 
 #include <unistd.h>
 
-#include <algorithm>
 #include <cerrno>
-#include <cstdlib>
 #include <cstring>
 #include <cwchar>
+#include <deque>
 #include <memory>
 #include <string>
 
+#include "builtins/abbr.h"
 #include "builtins/argparse.h"
 #include "builtins/bg.h"
 #include "builtins/bind.h"
@@ -50,6 +50,7 @@
 #include "builtins/history.h"
 #include "builtins/jobs.h"
 #include "builtins/math.h"
+#include "builtins/path.h"
 #include "builtins/printf.h"
 #include "builtins/pwd.h"
 #include "builtins/random.h"
@@ -65,13 +66,11 @@
 #include "builtins/type.h"
 #include "builtins/ulimit.h"
 #include "builtins/wait.h"
-#include "common.h"
 #include "complete.h"
-#include "exec.h"
 #include "fallback.h"  // IWYU pragma: keep
 #include "flog.h"
-#include "intern.h"
 #include "io.h"
+#include "null_terminated_array.h"
 #include "parse_constants.h"
 #include "parse_util.h"
 #include "parser.h"
@@ -107,8 +106,7 @@ void builtin_wperror(const wchar_t *program_name, io_streams_t &streams) {
 }
 
 static const wchar_t *const short_options = L"+:h";
-static const struct woption long_options[] = {{L"help", no_argument, nullptr, 'h'},
-                                              {nullptr, 0, nullptr, 0}};
+static const struct woption long_options[] = {{L"help", no_argument, 'h'}, {}};
 
 int parse_help_only_cmd_opts(struct help_only_cmd_opts_t &opts, int *optind, int argc,
                              const wchar_t **argv, parser_t &parser, io_streams_t &streams) {
@@ -149,11 +147,11 @@ void builtin_print_help(parser_t &parser, const io_streams_t &streams, const wch
                         wcstring *error_message) {
     // This won't ever work if no_exec is set.
     if (no_exec()) return;
-    const wcstring name_esc = escape_string(name, ESCAPE_ALL);
+    const wcstring name_esc = escape_string(name);
     wcstring cmd = format_string(L"__fish_print_help %ls ", name_esc.c_str());
     io_chain_t ios;
     if (error_message) {
-        cmd.append(escape_string(*error_message, ESCAPE_ALL));
+        cmd.append(escape_string(*error_message));
         // If it's an error, redirect the output of __fish_print_help to stderr
         ios.push_back(std::make_shared<io_fd_t>(STDOUT_FILENO, STDERR_FILENO));
     }
@@ -200,7 +198,7 @@ void builtin_print_error_trailer(parser_t &parser, output_stream_t &b, const wch
     b.append_format(_(L"(Type 'help %ls' for related documentation)\n"), cmd);
 }
 
-/// A generic bultin that only supports showing a help message. This is only a placeholder that
+/// A generic builtin that only supports showing a help message. This is only a placeholder that
 /// prints the help message. Useful for commands that live in the parser.
 static maybe_t<int> builtin_generic(parser_t &parser, io_streams_t &streams, const wchar_t **argv) {
     const wchar_t *cmd = argv[0];
@@ -339,7 +337,6 @@ static maybe_t<int> builtin_false(parser_t &parser, io_streams_t &streams, const
 
 static maybe_t<int> builtin_gettext(parser_t &parser, io_streams_t &streams, const wchar_t **argv) {
     UNUSED(parser);
-    UNUSED(streams);
     for (int i = 1; i < builtin_count_args(argv); i++) {
         streams.out.append(_(argv[i]));
     }
@@ -358,6 +355,7 @@ static constexpr builtin_data_t builtin_datas[] = {
     {L":", &builtin_true, N_(L"Return a successful result")},
     {L"[", &builtin_test, N_(L"Test a condition")},
     {L"_", &builtin_gettext, N_(L"Translate a string")},
+    {L"abbr", &builtin_abbr, N_(L"Manage abbreviations")},
     {L"and", &builtin_generic, N_(L"Run command if last command succeeded")},
     {L"argparse", &builtin_argparse, N_(L"Parse options in fish script")},
     {L"begin", &builtin_generic, N_(L"Create a block of code")},
@@ -394,6 +392,7 @@ static constexpr builtin_data_t builtin_datas[] = {
     {L"math", &builtin_math, N_(L"Evaluate math expressions")},
     {L"not", &builtin_generic, N_(L"Negate exit status of job")},
     {L"or", &builtin_generic, N_(L"Execute command if previous command failed")},
+    {L"path", &builtin_path, N_(L"Handle paths")},
     {L"printf", &builtin_printf, N_(L"Prints formatted text")},
     {L"pwd", &builtin_pwd, N_(L"Print the working directory")},
     {L"random", &builtin_random, N_(L"Generate random number")},
@@ -430,13 +429,6 @@ static const builtin_data_t *builtin_lookup(const wcstring &name) {
     return get_by_sorted_name(name.c_str(), builtin_datas);
 }
 
-/// Initialize builtin data.
-void builtin_init() {
-    for (const auto &builtin_data : builtin_datas) {
-        intern_static(builtin_data.name);
-    }
-}
-
 /// Is there a builtin command with the given name?
 bool builtin_exists(const wcstring &cmd) { return static_cast<bool>(builtin_lookup(cmd)); }
 
@@ -470,7 +462,7 @@ proc_status_t builtin_run(parser_t &parser, const wcstring_list_t &argv, io_stre
         // Resolve our status code.
         // If the builtin itself produced an error, use that error.
         // Otherwise use any errors from writing to out and writing to err, in that order.
-        int code = builtin_ret ? *builtin_ret : 0;
+        int code = builtin_ret.has_value() ? *builtin_ret : 0;
         if (code == 0) code = out_ret;
         if (code == 0) code = err_ret;
 
@@ -482,6 +474,9 @@ proc_status_t builtin_run(parser_t &parser, const wcstring_list_t &argv, io_stre
         if (code == 0 && !builtin_ret.has_value()) {
             return proc_status_t::empty();
         }
+        if (code < 0) {
+            FLOGF(warning, "builtin %ls returned invalid exit code %d", cmdname.c_str(), code);
+        }
         return proc_status_t::from_exit_code(code);
     }
 
