@@ -2,13 +2,12 @@
 #ifndef FISH_BUILTIN_H
 #define FISH_BUILTIN_H
 
-#include <stddef.h>
-
 #include <vector>
 
 #include "common.h"
+#include "complete.h"
+#include "maybe.h"
 
-class completion_t;
 class parser_t;
 class proc_status_t;
 class output_stream_t;
@@ -60,10 +59,10 @@ struct builtin_data_t {
 #define BUILTIN_ERR_MAX_ARG_COUNT1 _(L"%ls: expected <= %d arguments; got %d\n")
 
 /// Error message for invalid variable name.
-#define BUILTIN_ERR_VARNAME _(L"%ls: %ls: invalid variable. See `help identifiers`\n")
+#define BUILTIN_ERR_VARNAME _(L"%ls: %ls: invalid variable name. See `help identifiers`\n")
 
 /// Error message for invalid bind mode name.
-#define BUILTIN_ERR_BIND_MODE _(L"%ls: %ls: invalid mode. See `help identifiers`\n")
+#define BUILTIN_ERR_BIND_MODE _(L"%ls: %ls: invalid mode name. See `help identifiers`\n")
 
 /// Error message when too many arguments are supplied to a builtin.
 #define BUILTIN_ERR_TOO_MANY_ARGUMENTS _(L"%ls: too many arguments\n")
@@ -78,7 +77,6 @@ struct builtin_data_t {
 /// The send stuff to foreground message.
 #define FG_MSG _(L"Send job %d (%ls) to foreground\n")
 
-void builtin_init();
 bool builtin_exists(const wcstring &cmd);
 
 proc_status_t builtin_run(parser_t &parser, const wcstring_list_t &argv, io_streams_t &streams);
