@@ -3,8 +3,10 @@
 
 #include "bg.h"
 
+#include <sys/types.h>
+
 #include <cerrno>
-#include <cstdlib>
+#include <deque>
 #include <memory>
 #include <vector>
 
@@ -13,6 +15,7 @@
 #include "../fallback.h"  // IWYU pragma: keep
 #include "../io.h"
 #include "../job_group.h"
+#include "../maybe.h"
 #include "../parser.h"
 #include "../proc.h"
 #include "../wutil.h"  // IWYU pragma: keep
