@@ -52,21 +52,23 @@
 
 #include "printf.h"
 
-#include <sys/types.h>
-
 #include <cerrno>
+#include <cinttypes>
 #include <climits>
-#include <clocale>
 #include <cstdarg>
-#include <cstddef>
 #include <cstdint>
-#include <cstdlib>
 #include <cstring>
 #include <cwchar>
 #include <cwctype>
+#include <locale>
+#ifdef HAVE_XLOCALE_H
+#include <xlocale.h>
+#endif
 
 #include "../builtin.h"
 #include "../common.h"
+#include "../io.h"
+#include "../maybe.h"
 #include "../wcstringutil.h"
 #include "../wutil.h"  // IWYU pragma: keep
 
@@ -83,6 +85,10 @@ struct builtin_printf_state_t {
     // Whether we should stop outputting. This gets set in the case of an error, and also with the
     // \c escape.
     bool early_exit;
+    // Our output buffer, so we don't write() constantly.
+    // Our strategy is simple:
+    // We print once per argument, and we flush the buffer before the error.
+    wcstring buff;
 
     explicit builtin_printf_state_t(io_streams_t &s)
         : streams(s), exit_code(0), early_exit(false) {}
@@ -108,104 +114,16 @@ struct builtin_printf_state_t {
 
 static bool is_octal_digit(wchar_t c) { return iswdigit(c) && c < L'8'; }
 
-static int hex_to_bin(const wchar_t &c) {
-    switch (c) {
-        case L'0': {
-            return 0;
-        }
-        case L'1': {
-            return 1;
-        }
-        case L'2': {
-            return 2;
-        }
-        case L'3': {
-            return 3;
-        }
-        case L'4': {
-            return 4;
-        }
-        case L'5': {
-            return 5;
-        }
-        case L'6': {
-            return 6;
-        }
-        case L'7': {
-            return 7;
-        }
-        case L'8': {
-            return 8;
-        }
-        case L'9': {
-            return 9;
-        }
-        case L'a':
-        case L'A': {
-            return 10;
-        }
-        case L'b':
-        case L'B': {
-            return 11;
-        }
-        case L'c':
-        case L'C': {
-            return 12;
-        }
-        case L'd':
-        case L'D': {
-            return 13;
-        }
-        case L'e':
-        case L'E': {
-            return 14;
-        }
-        case L'f':
-        case L'F': {
-            return 15;
-        }
-        default: {
-            return -1;
-        }
-    }
-}
-
-static int octal_to_bin(wchar_t c) {
-    switch (c) {
-        case L'0': {
-            return 0;
-        }
-        case L'1': {
-            return 1;
-        }
-        case L'2': {
-            return 2;
-        }
-        case L'3': {
-            return 3;
-        }
-        case L'4': {
-            return 4;
-        }
-        case L'5': {
-            return 5;
-        }
-        case L'6': {
-            return 6;
-        }
-        case L'7': {
-            return 7;
-        }
-        default: {
-            return -1;
-        }
-    }
-}
-
 void builtin_printf_state_t::nonfatal_error(const wchar_t *fmt, ...) {
     // Don't error twice.
     if (early_exit) return;
 
+    // If we have output, write it so it appears first.
+    if (!buff.empty()) {
+        streams.out.append(buff);
+        buff.clear();
+    }
+
     va_list va;
     va_start(va, fmt);
     wcstring errstr = vformat_string(fmt, va);
@@ -222,6 +140,12 @@ void builtin_printf_state_t::fatal_error(const wchar_t *fmt, ...) {
     // Don't error twice.
     if (early_exit) return;
 
+    // If we have output, write it so it appears first.
+    if (!buff.empty()) {
+        streams.out.append(buff);
+        buff.clear();
+    }
+
     va_list va;
     va_start(va, fmt);
     wcstring errstr = vformat_string(fmt, va);
@@ -236,7 +160,7 @@ void builtin_printf_state_t::append_output(wchar_t c) {
     // Don't output if we're done.
     if (early_exit) return;
 
-    streams.out.push_back(c);
+    buff.push_back(c);
 }
 
 void builtin_printf_state_t::append_format_output(const wchar_t *fmt, ...) {
@@ -247,11 +171,11 @@ void builtin_printf_state_t::append_format_output(const wchar_t *fmt, ...) {
     va_start(va, fmt);
     wcstring tmp = vformat_string(fmt, va);
     va_end(va);
-    streams.out.append(tmp);
+    buff.append(tmp);
 }
 
 void builtin_printf_state_t::verify_numeric(const wchar_t *s, const wchar_t *end, int errcode) {
-    if (errcode != 0) {
+    if (errcode != 0 && errcode != EINVAL) {
         if (errcode == ERANGE) {
             this->fatal_error(L"%ls: %ls", s, _(L"Number out of range"));
         } else {
@@ -262,7 +186,15 @@ void builtin_printf_state_t::verify_numeric(const wchar_t *s, const wchar_t *end
             this->fatal_error(_(L"%ls: expected a numeric value"), s);
         } else {
             // This isn't entirely fatal - the value should still be printed.
-            this->nonfatal_error(_(L"%ls: value not completely converted (can't convert '%ls')"), s, end);
+            this->nonfatal_error(_(L"%ls: value not completely converted (can't convert '%ls')"), s,
+                                 end);
+            // Warn about octal numbers as they can be confusing.
+            // Do it if the unconverted digit is a valid hex digit,
+            // because it could also be an "0x" -> "0" typo.
+            if (*s == L'0' && iswxdigit(*end)) {
+                this->nonfatal_error(
+                    _(L"Hint: a leading '0' without an 'x' indicates an octal number"), s, end);
+            }
         }
     }
 }
@@ -270,16 +202,14 @@ void builtin_printf_state_t::verify_numeric(const wchar_t *s, const wchar_t *end
 template <typename T>
 static T raw_string_to_scalar_type(const wchar_t *s, wchar_t **end);
 
-// we use wcstoll instead of wcstoimax because FreeBSD 8 has busted wcstoumax and wcstoimax - see
-// #626
 template <>
 intmax_t raw_string_to_scalar_type(const wchar_t *s, wchar_t **end) {
-    return std::wcstoll(s, end, 0);
+    return std::wcstoimax(s, end, 0);
 }
 
 template <>
 uintmax_t raw_string_to_scalar_type(const wchar_t *s, wchar_t **end) {
-    return std::wcstoull(s, end, 0);
+    return std::wcstoumax(s, end, 0);
 }
 
 template <>
@@ -366,7 +296,7 @@ long builtin_printf_state_t::print_esc(const wchar_t *escstart, bool octal_0) {
     if (*p == L'x') {
         // A hexadecimal \xhh escape sequence must have 1 or 2 hex. digits.
         for (esc_length = 0, ++p; esc_length < 2 && iswxdigit(*p); ++esc_length, ++p)
-            esc_value = esc_value * 16 + hex_to_bin(*p);
+            esc_value = esc_value * 16 + convert_digit(*p, 16);
         if (esc_length == 0) this->fatal_error(_(L"missing hexadecimal number in escape"));
         this->append_output(ENCODE_DIRECT_BASE + esc_value % 256);
     } else if (is_octal_digit(*p)) {
@@ -375,7 +305,7 @@ long builtin_printf_state_t::print_esc(const wchar_t *escstart, bool octal_0) {
         // Wrap mod 256, which matches historic behavior.
         for (esc_length = 0, p += octal_0 && *p == L'0'; esc_length < 3 && is_octal_digit(*p);
              ++esc_length, ++p)
-            esc_value = esc_value * 8 + octal_to_bin(*p);
+            esc_value = esc_value * 8 + convert_digit(*p, 8);
         this->append_output(ENCODE_DIRECT_BASE + esc_value % 256);
     } else if (*p && std::wcschr(L"\"\\abcefnrtv", *p)) {
         print_esc_char(*p++);
@@ -391,7 +321,7 @@ long builtin_printf_state_t::print_esc(const wchar_t *escstart, bool octal_0) {
                 }
                 break;
             }
-            uni_value = uni_value * 16 + hex_to_bin(*p);
+            uni_value = uni_value * 16 + convert_digit(*p, 16);
             p++;
         }
 
@@ -733,21 +663,12 @@ int builtin_printf_state_t::print_formatted(const wchar_t *format, int argc, con
 
 /// The printf builtin.
 maybe_t<int> builtin_printf(parser_t &parser, io_streams_t &streams, const wchar_t **argv) {
-    const wchar_t *cmd = argv[0];
+    UNUSED(parser);
     int argc = builtin_count_args(argv);
-    help_only_cmd_opts_t opts;
 
-    int optind;
-    int retval = parse_help_only_cmd_opts(opts, &optind, argc, argv, parser, streams);
-    if (retval != STATUS_CMD_OK) return retval;
+    argv++;
+    argc--;
 
-    if (opts.print_help) {
-        builtin_print_help(parser, streams, cmd);
-        return STATUS_CMD_OK;
-    }
-
-    argc -= optind;
-    argv += optind;
     if (argc < 1) {
         return STATUS_INVALID_ARGS;
     }
@@ -776,6 +697,10 @@ maybe_t<int> builtin_printf(parser_t &parser, io_streams_t &streams, const wchar
         args_used = state.print_formatted(format, argc, argv);
         argc -= args_used;
         argv += args_used;
+        if (!state.buff.empty()) {
+            streams.out.append(state.buff);
+            state.buff.clear();
+        }
     } while (args_used > 0 && argc > 0 && !state.early_exit);
 
 #if defined(HAVE_USELOCALE) || defined(__GLIBC__)
