@@ -3,34 +3,26 @@
 
 #include "read.h"
 
-#include <termios.h>
 #include <unistd.h>
 
 #include <algorithm>
 #include <cerrno>
-#include <climits>
-#include <cstddef>
 #include <cstdio>
 #include <cstdlib>
 #include <cstring>
 #include <cwchar>
-#include <memory>
-#include <numeric>
 #include <string>
-#include <vector>
+#include <utility>
 
 #include "../builtin.h"
 #include "../common.h"
-#include "../complete.h"
 #include "../env.h"
-#include "../event.h"
 #include "../fallback.h"  // IWYU pragma: keep
-#include "../highlight.h"
-#include "../history.h"
 #include "../io.h"
+#include "../maybe.h"
 #include "../parser.h"
-#include "../proc.h"
 #include "../reader.h"
+#include "../tokenizer.h"
 #include "../wcstringutil.h"
 #include "../wgetopt.h"
 #include "../wutil.h"  // IWYU pragma: keep
@@ -60,27 +52,27 @@ struct read_cmd_opts_t {
 }  // namespace
 
 static const wchar_t *const short_options = L":ac:d:fghiLln:p:sStuxzP:UR:L";
-static const struct woption long_options[] = {{L"array", no_argument, nullptr, 'a'},
-                                              {L"command", required_argument, nullptr, 'c'},
-                                              {L"delimiter", required_argument, nullptr, 'd'},
-                                              {L"export", no_argument, nullptr, 'x'},
-                                              {L"function", no_argument, nullptr, 'f'},
-                                              {L"global", no_argument, nullptr, 'g'},
-                                              {L"help", no_argument, nullptr, 'h'},
-                                              {L"line", no_argument, nullptr, 'L'},
-                                              {L"list", no_argument, nullptr, 'a'},
-                                              {L"local", no_argument, nullptr, 'l'},
-                                              {L"nchars", required_argument, nullptr, 'n'},
-                                              {L"null", no_argument, nullptr, 'z'},
-                                              {L"prompt", required_argument, nullptr, 'p'},
-                                              {L"prompt-str", required_argument, nullptr, 'P'},
-                                              {L"right-prompt", required_argument, nullptr, 'R'},
-                                              {L"shell", no_argument, nullptr, 'S'},
-                                              {L"silent", no_argument, nullptr, 's'},
-                                              {L"tokenize", no_argument, nullptr, 't'},
-                                              {L"unexport", no_argument, nullptr, 'u'},
-                                              {L"universal", no_argument, nullptr, 'U'},
-                                              {nullptr, 0, nullptr, 0}};
+static const struct woption long_options[] = {{L"array", no_argument, 'a'},
+                                              {L"command", required_argument, 'c'},
+                                              {L"delimiter", required_argument, 'd'},
+                                              {L"export", no_argument, 'x'},
+                                              {L"function", no_argument, 'f'},
+                                              {L"global", no_argument, 'g'},
+                                              {L"help", no_argument, 'h'},
+                                              {L"line", no_argument, 'L'},
+                                              {L"list", no_argument, 'a'},
+                                              {L"local", no_argument, 'l'},
+                                              {L"nchars", required_argument, 'n'},
+                                              {L"null", no_argument, 'z'},
+                                              {L"prompt", required_argument, 'p'},
+                                              {L"prompt-str", required_argument, 'P'},
+                                              {L"right-prompt", required_argument, 'R'},
+                                              {L"shell", no_argument, 'S'},
+                                              {L"silent", no_argument, 's'},
+                                              {L"tokenize", no_argument, 't'},
+                                              {L"unexport", no_argument, 'u'},
+                                              {L"universal", no_argument, 'U'},
+                                              {}};
 
 static int parse_cmd_opts(read_cmd_opts_t &opts, int *optind,  //!OCLINT(high ncss method)
                           int argc, const wchar_t **argv, parser_t &parser, io_streams_t &streams) {
@@ -224,6 +216,7 @@ static int read_interactive(parser_t &parser, wcstring &buff, int nchars, bool s
 
     conf.left_prompt_cmd = prompt;
     conf.right_prompt_cmd = right_prompt;
+    conf.event = L"fish_read";
 
     conf.in = in;
 
@@ -233,7 +226,6 @@ static int read_interactive(parser_t &parser, wcstring &buff, int nchars, bool s
     commandline_set_buffer(commandline, std::wcslen(commandline));
     scoped_push<bool> interactive{&parser.libdata().is_interactive, true};
 
-    event_fire_generic(parser, L"fish_read");
     auto mline = reader_readline(nchars);
     interactive.restore();
     if (mline) {
@@ -262,7 +254,7 @@ static int read_interactive(parser_t &parser, wcstring &buff, int nchars, bool s
 /// of chars.
 ///
 /// Returns an exit status.
-static int read_in_chunks(int fd, wcstring &buff, bool split_null) {
+static int read_in_chunks(int fd, wcstring &buff, bool split_null, bool do_seek) {
     int exit_res = STATUS_CMD_OK;
     std::string str;
     bool eof = false;
@@ -284,7 +276,7 @@ static int read_in_chunks(int fd, wcstring &buff, bool split_null) {
         if (bytes_consumed < bytes_read) {
             // We found a splitter. The +1 because we need to treat the splitter as consumed, but
             // not append it to the string.
-            if (lseek(fd, bytes_consumed - bytes_read + 1, SEEK_CUR) == -1) {
+            if (do_seek && lseek(fd, bytes_consumed - bytes_read + 1, SEEK_CUR) == -1) {
                 wperror(L"lseek");
                 return STATUS_CMD_ERROR;
             }
@@ -380,7 +372,7 @@ static int validate_read_args(const wchar_t *cmd, read_cmd_opts_t &opts, int arg
     }
 
     if (opts.prompt_str) {
-        opts.prompt_cmd = L"echo " + escape_string(opts.prompt_str, ESCAPE_ALL);
+        opts.prompt_cmd = L"echo " + escape_string(opts.prompt_str);
         opts.prompt = opts.prompt_cmd.c_str();
     } else if (!opts.prompt) {
         opts.prompt = DEFAULT_READ_PROMPT;
@@ -494,8 +486,6 @@ maybe_t<int> builtin_read(parser_t &parser, io_streams_t &streams, const wchar_t
     do {
         buff.clear();
 
-        // TODO: Determine if the original set of conditions for interactive reads should be
-        // reinstated: if (isatty(0) && streams.stdin_fd == STDIN_FILENO && !split_null) {
         int stream_stdin_is_a_tty = isatty(streams.stdin_fd);
         if (stream_stdin_is_a_tty && !opts.split_null) {
             // Read interactively using reader_readline(). This does not support splitting on null.
@@ -503,8 +493,21 @@ maybe_t<int> builtin_read(parser_t &parser, io_streams_t &streams, const wchar_t
                 read_interactive(parser, buff, opts.nchars, opts.shell, opts.silent, opts.prompt,
                                  opts.right_prompt, opts.commandline, streams.stdin_fd);
         } else if (!opts.nchars && !stream_stdin_is_a_tty &&
-                   lseek(streams.stdin_fd, 0, SEEK_CUR) != -1) {
-            exit_res = read_in_chunks(streams.stdin_fd, buff, opts.split_null);
+                   // "one_line" is implemented as reading n-times to a new line,
+                   // if we're chunking we could get multiple lines so we would have to advance
+                   // more than 1 per run through the loop. Let's skip that for now.
+                   !opts.one_line &&
+                   (streams.stdin_is_directly_redirected ||
+                    lseek(streams.stdin_fd, 0, SEEK_CUR) != -1)) {
+            // We read in chunks when we either can seek (so we put the bytes back),
+            // or we have the bytes to ourselves (because it's directly redirected).
+            //
+            // Note we skip seeking back even if we're directly redirected to a seekable stream,
+            // under the assumption that the stream will be closed soon anyway.
+            // You don't rewind VHS tapes before throwing them in the trash.
+            // TODO: Do this when nchars is set by seeking back.
+            exit_res = read_in_chunks(streams.stdin_fd, buff, opts.split_null,
+                                      !streams.stdin_is_directly_redirected);
         } else {
             exit_res =
                 read_one_char_at_a_time(streams.stdin_fd, buff, opts.nchars, opts.split_null);
