@@ -3,11 +3,12 @@
 
 #include "set_color.h"
 
-#include <cstddef>
+#include <unistd.h>
+
 #include <cstdlib>
 
 #if HAVE_CURSES_H
-#include <curses.h>
+#include <curses.h>  // IWYU pragma: keep
 #elif HAVE_NCURSES_H
 #include <ncurses.h>
 #elif HAVE_NCURSES_CURSES_H
@@ -19,7 +20,6 @@
 #include <ncurses/term.h>
 #endif
 
-#include <memory>
 #include <string>
 #include <vector>
 
@@ -27,8 +27,10 @@
 #include "../color.h"
 #include "../common.h"
 #include "../env.h"
+#include "../fallback.h"  // IWYU pragma: keep
+#include "../io.h"
+#include "../maybe.h"
 #include "../output.h"
-#include "../parser.h"
 #include "../wgetopt.h"
 #include "../wutil.h"  // IWYU pragma: keep
 
@@ -63,10 +65,11 @@ static void print_modifiers(outputter_t &outp, bool bold, bool underline, bool i
     }
 }
 
-static void print_colors(io_streams_t &streams, bool bold, bool underline, bool italics, bool dim,
-                         bool reverse, rgb_color_t bg) {
+static void print_colors(io_streams_t &streams, wcstring_list_t args, bool bold, bool underline,
+                         bool italics, bool dim, bool reverse, rgb_color_t bg) {
     outputter_t outp;
-    for (const auto &color_name : rgb_color_t::named_color_names()) {
+    if (args.empty()) args = rgb_color_t::named_color_names();
+    for (const auto &color_name : args) {
         if (!streams.out_is_redirected && isatty(STDOUT_FILENO)) {
             print_modifiers(outp, bold, underline, italics, dim, reverse, bg);
             rgb_color_t color = rgb_color_t(color_name);
@@ -88,43 +91,22 @@ static void print_colors(io_streams_t &streams, bool bold, bool underline, bool
 }
 
 static const wchar_t *const short_options = L":b:hvoidrcu";
-static const struct woption long_options[] = {{L"background", required_argument, nullptr, 'b'},
-                                              {L"help", no_argument, nullptr, 'h'},
-                                              {L"bold", no_argument, nullptr, 'o'},
-                                              {L"underline", no_argument, nullptr, 'u'},
-                                              {L"italics", no_argument, nullptr, 'i'},
-                                              {L"dim", no_argument, nullptr, 'd'},
-                                              {L"reverse", no_argument, nullptr, 'r'},
-                                              {L"version", no_argument, nullptr, 'v'},
-                                              {L"print-colors", no_argument, nullptr, 'c'},
-                                              {nullptr, 0, nullptr, 0}};
-
-#ifdef __APPLE__
-static char sitm_esc[] = "\x1B[3m";
-static char ritm_esc[] = "\x1B[23m";
-static char dim_esc[] = "\x1B[2m";
-#endif
+static const struct woption long_options[] = {{L"background", required_argument, 'b'},
+                                              {L"help", no_argument, 'h'},
+                                              {L"bold", no_argument, 'o'},
+                                              {L"underline", no_argument, 'u'},
+                                              {L"italics", no_argument, 'i'},
+                                              {L"dim", no_argument, 'd'},
+                                              {L"reverse", no_argument, 'r'},
+                                              {L"version", no_argument, 'v'},
+                                              {L"print-colors", no_argument, 'c'},
+                                              {}};
 
 /// set_color builtin.
 maybe_t<int> builtin_set_color(parser_t &parser, io_streams_t &streams, const wchar_t **argv) {
     // By the time this is called we should have initialized the curses subsystem.
     assert(curses_initialized);
 
-// Hack in missing italics and dim capabilities omitted from MacOS xterm-256color terminfo
-// Helps Terminal.app/iTerm
-#ifdef __APPLE__
-    const auto term_prog = parser.vars().get(L"TERM_PROGRAM");
-    if (!term_prog.missing_or_empty() &&
-        (term_prog->as_string() == L"Apple_Terminal" || term_prog->as_string() == L"iTerm.app")) {
-        const auto term = parser.vars().get(L"TERM");
-        if (!term.missing_or_empty() && (term->as_string() == L"xterm-256color")) {
-            enter_italics_mode = sitm_esc;
-            exit_italics_mode = ritm_esc;
-            enter_dim_mode = dim_esc;
-        }
-    }
-#endif
-
     // Variables used for parsing the argument list.
     int argc = builtin_count_args(argv);
 
@@ -203,7 +185,8 @@ maybe_t<int> builtin_set_color(parser_t &parser, io_streams_t &streams, const wc
         if (bgcolor && bg.is_special()) {
             bg = rgb_color_t(L"");
         }
-        print_colors(streams, bold, underline, italics, dim, reverse, bg);
+        wcstring_list_t args(argv + w.woptind, argv + argc);
+        print_colors(streams, args, bold, underline, italics, dim, reverse, bg);
         return STATUS_CMD_OK;
     }
 
