@@ -3,8 +3,14 @@
 
 #include "status.h"
 
-#include <cstddef>
+#include <unistd.h>
+
+#include <algorithm>
+#include <cerrno>
+#include <cstring>
 #include <cwchar>
+#include <limits>
+#include <memory>
 #include <string>
 
 #include "../builtin.h"
@@ -13,6 +19,7 @@
 #include "../fallback.h"  // IWYU pragma: keep
 #include "../future_feature_flags.h"
 #include "../io.h"
+#include "../maybe.h"
 #include "../parser.h"
 #include "../proc.h"
 #include "../wgetopt.h"
@@ -38,6 +45,7 @@ enum status_cmd_t {
     STATUS_SET_JOB_CONTROL,
     STATUS_STACK_TRACE,
     STATUS_TEST_FEATURE,
+    STATUS_CURRENT_COMMANDLINE,
     STATUS_UNDEF
 };
 
@@ -46,6 +54,7 @@ const enum_map<status_cmd_t> status_enum_map[] = {
     {STATUS_BASENAME, L"basename"},
     {STATUS_BASENAME, L"current-basename"},
     {STATUS_CURRENT_CMD, L"current-command"},
+    {STATUS_CURRENT_COMMANDLINE, L"current-commandline"},
     {STATUS_DIRNAME, L"current-dirname"},
     {STATUS_FILENAME, L"current-filename"},
     {STATUS_FUNCTION, L"current-function"},
@@ -100,7 +109,6 @@ namespace {
 struct status_cmd_opts_t {
     int level{1};
     maybe_t<job_control_t> new_job_control_mode{};
-    const wchar_t *feature_name{};
     status_cmd_t status_cmd{STATUS_UNDEF};
     bool print_help{false};
 };
@@ -112,24 +120,24 @@ struct status_cmd_opts_t {
 /// scripts.
 static const wchar_t *const short_options = L":L:cbilfnhj:t";
 static const struct woption long_options[] = {
-    {L"help", no_argument, nullptr, 'h'},
-    {L"current-filename", no_argument, nullptr, 'f'},
-    {L"current-line-number", no_argument, nullptr, 'n'},
-    {L"filename", no_argument, nullptr, 'f'},
-    {L"fish-path", no_argument, nullptr, STATUS_FISH_PATH},
-    {L"is-block", no_argument, nullptr, 'b'},
-    {L"is-command-substitution", no_argument, nullptr, 'c'},
-    {L"is-full-job-control", no_argument, nullptr, STATUS_IS_FULL_JOB_CTRL},
-    {L"is-interactive", no_argument, nullptr, 'i'},
-    {L"is-interactive-job-control", no_argument, nullptr, STATUS_IS_INTERACTIVE_JOB_CTRL},
-    {L"is-login", no_argument, nullptr, 'l'},
-    {L"is-no-job-control", no_argument, nullptr, STATUS_IS_NO_JOB_CTRL},
-    {L"job-control", required_argument, nullptr, 'j'},
-    {L"level", required_argument, nullptr, 'L'},
-    {L"line", no_argument, nullptr, 'n'},
-    {L"line-number", no_argument, nullptr, 'n'},
-    {L"print-stack-trace", no_argument, nullptr, 't'},
-    {nullptr, 0, nullptr, 0}};
+    {L"help", no_argument, 'h'},
+    {L"current-filename", no_argument, 'f'},
+    {L"current-line-number", no_argument, 'n'},
+    {L"filename", no_argument, 'f'},
+    {L"fish-path", no_argument, STATUS_FISH_PATH},
+    {L"is-block", no_argument, 'b'},
+    {L"is-command-substitution", no_argument, 'c'},
+    {L"is-full-job-control", no_argument, STATUS_IS_FULL_JOB_CTRL},
+    {L"is-interactive", no_argument, 'i'},
+    {L"is-interactive-job-control", no_argument, STATUS_IS_INTERACTIVE_JOB_CTRL},
+    {L"is-login", no_argument, 'l'},
+    {L"is-no-job-control", no_argument, STATUS_IS_NO_JOB_CTRL},
+    {L"job-control", required_argument, 'j'},
+    {L"level", required_argument, 'L'},
+    {L"line", no_argument, 'n'},
+    {L"line-number", no_argument, 'n'},
+    {L"print-stack-trace", no_argument, 't'},
+    {}};
 
 /// Remember the status subcommand and disallow selecting more than one status subcommand.
 static bool set_status_cmd(const wchar_t *cmd, status_cmd_opts_t &opts, status_cmd_t sub_cmd,
@@ -370,7 +378,7 @@ maybe_t<int> builtin_status(parser_t &parser, io_streams_t &streams, const wchar
         case STATUS_FILENAME: {
             CHECK_FOR_UNEXPECTED_STATUS_ARGS(opts.status_cmd)
             auto res = parser.current_filename();
-            wcstring fn = res ? res : L"";
+            wcstring fn = res ? *res : L"";
             if (!fn.empty() && opts.status_cmd == STATUS_DIRNAME) {
                 fn = wdirname(fn);
             } else if (!fn.empty() && opts.status_cmd == STATUS_BASENAME) {
@@ -442,10 +450,9 @@ maybe_t<int> builtin_status(parser_t &parser, io_streams_t &streams, const wchar
         }
         case STATUS_CURRENT_CMD: {
             CHECK_FOR_UNEXPECTED_STATUS_ARGS(opts.status_cmd)
-            // HACK: Go via the deprecated variable to get the command.
-            const auto var = parser.vars().get(L"_");
-            if (!var.missing_or_empty()) {
-                streams.out.append(var->as_string());
+            const auto &var = parser.libdata().status_vars.command;
+            if (!var.empty()) {
+                streams.out.append(var);
                 streams.out.push_back(L'\n');
             } else {
                 streams.out.append(program_name);
@@ -453,12 +460,43 @@ maybe_t<int> builtin_status(parser_t &parser, io_streams_t &streams, const wchar
             }
             break;
         }
-        case STATUS_FISH_PATH: {
+        case STATUS_CURRENT_COMMANDLINE: {
             CHECK_FOR_UNEXPECTED_STATUS_ARGS(opts.status_cmd)
-            streams.out.append(str2wcstring(get_executable_path("fish")));
+            const auto &var = parser.libdata().status_vars.commandline;
+            streams.out.append(var);
             streams.out.push_back(L'\n');
             break;
         }
+        case STATUS_FISH_PATH: {
+            CHECK_FOR_UNEXPECTED_STATUS_ARGS(opts.status_cmd);
+            auto path = str2wcstring(get_executable_path("fish"));
+            if (path.empty()) {
+                streams.err.append_format(L"%ls: Could not get executable path: '%s'\n", cmd,
+                                          std::strerror(errno));
+                break;
+            }
+
+            if (path[0] == L'/') {
+                // This is an absolute path, we can canonicalize it.
+                auto real = wrealpath(path);
+                if (real && waccess(*real, F_OK)) {
+                    streams.out.append(*real);
+                    streams.out.push_back(L'\n');
+                } else {
+                    // realpath did not work, just append the path
+                    // - maybe this was obtained via $PATH?
+                    streams.out.append(path);
+                    streams.out.push_back(L'\n');
+                }
+            } else {
+                // This is a relative path, it depends on where fish's parent process
+                // was when it started it and its idea of $PATH.
+                // The best we can do is to print it directly and hope it works.
+                streams.out.append(path);
+                streams.out.push_back(L'\n');
+            }
+            break;
+        }
     }
 
     return retval;
