@@ -10,23 +10,26 @@
 #include <unistd.h>
 
 #include <cerrno>
+#include <climits>
 #include <cmath>
 #include <cstdarg>
 #include <cstring>
 #include <cwchar>
 #include <cwctype>
+#include <map>
 #include <memory>
 #include <string>
-#include <type_traits>
 #include <utility>
+#include <vector>
 
 #include "../builtin.h"
 #include "../common.h"
+#include "../fallback.h"  // IWYU pragma: keep
 #include "../io.h"
+#include "../maybe.h"
 #include "../parser.h"
 #include "../wutil.h"  // IWYU pragma: keep
 
-using std::move;
 using std::unique_ptr;
 
 namespace {
@@ -65,6 +68,10 @@ enum token_t {
     test_string_equal,      // "=", true if strings are identical
     test_string_not_equal,  // "!=", true if strings are not identical
 
+    test_file_newer,  // f1 -nt f2, true if f1 exists and is newer than f2, or there is no f2
+    test_file_older,  // f1 -ot f2, true if f2 exists and f1 does not, or f1 is older than f2
+    test_file_same,   // f1 -ef f2, true if f1 and f2 exist and refer to same file
+
     test_number_equal,          // "-eq", true if numbers are equal
     test_number_not_equal,      // "-ne", true if numbers are not equal
     test_number_greater,        // "-gt", true if first number is larger than second
@@ -152,6 +159,9 @@ static const token_info_t *token_for_string(const wcstring &str) {
         {L"-z", {test_string_z, UNARY_PRIMARY}},
         {L"=", {test_string_equal, BINARY_PRIMARY}},
         {L"!=", {test_string_not_equal, BINARY_PRIMARY}},
+        {L"-nt", {test_file_newer, BINARY_PRIMARY}},
+        {L"-ot", {test_file_older, BINARY_PRIMARY}},
+        {L"-ef", {test_file_same, BINARY_PRIMARY}},
         {L"-eq", {test_number_equal, BINARY_PRIMARY}},
         {L"-ne", {test_number_not_equal, BINARY_PRIMARY}},
         {L"-gt", {test_number_greater, BINARY_PRIMARY}},
@@ -260,7 +270,7 @@ class unary_operator final : public expression {
    public:
     unique_ptr<expression> subject;
     unary_operator(token_t tok, range_t where, unique_ptr<expression> exp)
-        : expression(tok, where), subject(move(exp)) {}
+        : expression(tok, where), subject(std::move(exp)) {}
     bool evaluate(io_streams_t *streams, wcstring_list_t &errors) override;
 };
 
@@ -288,7 +298,7 @@ class parenthetical_expression final : public expression {
    public:
     unique_ptr<expression> contents;
     parenthetical_expression(token_t tok, range_t where, unique_ptr<expression> expr)
-        : expression(tok, where), contents(move(expr)) {}
+        : expression(tok, where), contents(std::move(expr)) {}
 
     bool evaluate(io_streams_t *streams, wcstring_list_t &errors) override;
 };
@@ -325,7 +335,7 @@ unique_ptr<expression> test_parser::parse_unary_expression(unsigned int start, u
         unique_ptr<expression> subject(parse_unary_expression(start + 1, end));
         if (subject) {
             return make_unique<unary_operator>(tok, range_t(start, subject->range.end),
-                                               move(subject));
+                                               std::move(subject));
         }
         return nullptr;
     }
@@ -371,7 +381,7 @@ unique_ptr<expression> test_parser::parse_combining_expression(unsigned int star
 
         // Go to the end of this expression.
         idx = expr->range.end;
-        subjects.push_back(move(expr));
+        subjects.push_back(std::move(expr));
         first = false;
     }
 
@@ -380,8 +390,8 @@ unique_ptr<expression> test_parser::parse_combining_expression(unsigned int star
     }
     // Our new expression takes ownership of all expressions we created. The token we pass is
     // irrelevant.
-    return make_unique<combining_expression>(test_combine_and, range_t(start, idx), move(subjects),
-                                             move(combiners));
+    return make_unique<combining_expression>(test_combine_and, range_t(start, idx),
+                                             std::move(subjects), std::move(combiners));
 }
 
 unique_ptr<expression> test_parser::parse_unary_primary(unsigned int start, unsigned int end) {
@@ -460,7 +470,7 @@ unique_ptr<expression> test_parser::parse_parenthentical(unsigned int start, uns
 
     // Success.
     return make_unique<parenthetical_expression>(test_paren_open, range_t(start, close_index + 1),
-                                                 move(subexpr));
+                                                 std::move(subexpr));
 }
 
 unique_ptr<expression> test_parser::parse_primary(unsigned int start, unsigned int end) {
@@ -491,10 +501,10 @@ unique_ptr<expression> test_parser::parse_3_arg_expression(unsigned int start, u
             // Transfer ownership to the vector of subjects.
             std::vector<token_t> combiners = {center_token->tok};
             std::vector<unique_ptr<expression>> subjects;
-            subjects.push_back(move(left));
-            subjects.push_back(move(right));
+            subjects.push_back(std::move(left));
+            subjects.push_back(std::move(right));
             result = make_unique<combining_expression>(center_token->tok, range_t(start, end),
-                                                       move(subjects), move(combiners));
+                                                       std::move(subjects), std::move(combiners));
         }
     } else {
         result = parse_unary_expression(start, end);
@@ -511,7 +521,7 @@ unique_ptr<expression> test_parser::parse_4_arg_expression(unsigned int start, u
         unique_ptr<expression> subject(parse_3_arg_expression(start + 1, end));
         if (subject) {
             result = make_unique<unary_operator>(first_token, range_t(start, subject->range.end),
-                                                 move(subject));
+                                                 std::move(subject));
         }
     } else if (first_token == test_paren_open) {
         result = parse_parenthentical(start, end);
@@ -669,13 +679,14 @@ bool parenthetical_expression::evaluate(io_streams_t *streams, wcstring_list_t &
 }
 
 // Parse a double from arg. Return true on success, false on failure.
-static bool parse_double(const wchar_t *arg, double *out_res) {
+static bool parse_double(const wcstring &argstr, double *out_res) {
     // Consume leading spaces.
+    const wchar_t *arg = argstr.c_str();
     while (arg && *arg != L'\0' && iswspace(*arg)) arg++;
     if (!arg) return false;
     errno = 0;
     wchar_t *end = nullptr;
-    *out_res = fish_wcstod(arg, &end);
+    *out_res = fish_wcstod(arg, &end, argstr.size() - (arg - argstr.c_str()));
     // Consume trailing spaces.
     while (end && *end != L'\0' && iswspace(*end)) end++;
     return errno == 0 && end > arg && *end == L'\0';
@@ -688,7 +699,7 @@ static bool parse_double(const wchar_t *arg, double *out_res) {
 static bool parse_number(const wcstring &arg, number_t *number, wcstring_list_t &errors) {
     const wchar_t *argcs = arg.c_str();
     double floating = 0;
-    bool got_float = parse_double(argcs, &floating);
+    bool got_float = parse_double(arg, &floating);
     errno = 0;
     long long integral = fish_wcstoll(argcs);
     bool got_int = (errno == 0);
@@ -741,6 +752,15 @@ static bool binary_primary_evaluate(test_expressions::token_t token, const wcstr
         case test_string_not_equal: {
             return left != right;
         }
+        case test_file_newer: {
+            return file_id_for_path(right).older_than(file_id_for_path(left));
+        }
+        case test_file_older: {
+            return file_id_for_path(left).older_than(file_id_for_path(right));
+        }
+        case test_file_same: {
+            return file_id_for_path(left) == file_id_for_path(right);
+        }
         case test_number_equal: {
             return parse_number(left, &ln, errors) && parse_number(right, &rn, errors) &&
                    ln.compare(rn) == 0;
