@@ -10,21 +10,12 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <limits.h>
-#include <paths.h>
 #include <pthread.h>
 #include <stdarg.h>
-#include <stddef.h>
-#include <stdint.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <sys/stat.h>
 #include <sys/time.h>
 #include <termios.h>
 #include <unistd.h>
-#include <wctype.h>
 
-#include <cstring>
-#include <cwchar>
 #ifdef HAVE_EXECINFO_H
 #include <execinfo.h>
 #endif
@@ -35,32 +26,32 @@
 #endif
 
 #include <algorithm>
-#include <atomic>
-#include <memory>  // IWYU pragma: keep
-#include <type_traits>
+#include <csignal>
+#include <cstdlib>
+#include <cstring>
+#include <cwchar>
+#include <memory>
 
 #include "common.h"
-#include "env.h"
 #include "expand.h"
 #include "fallback.h"  // IWYU pragma: keep
 #include "flog.h"
 #include "future_feature_flags.h"
 #include "global_safety.h"
 #include "iothread.h"
-#include "parser.h"
-#include "proc.h"
 #include "signal.h"
 #include "termsize.h"
+#include "topic_monitor.h"
 #include "wcstringutil.h"
 #include "wildcard.h"
 #include "wutil.h"  // IWYU pragma: keep
 
 // Keep after "common.h"
 #ifdef HAVE_SYS_SYSCTL_H
-#include <sys/sysctl.h>
+#include <sys/sysctl.h>  // IWYU pragma: keep
 #endif
 #if defined(__APPLE__)
-#include <mach-o/dyld.h>
+#include <mach-o/dyld.h>  // IWYU pragma: keep
 #endif
 
 struct termios shell_modes;
@@ -95,8 +86,6 @@ const wchar_t *program_name;
 /// This is set during startup and not modified after.
 static relaxed_atomic_t<pid_t> initial_fg_process_group{-1};
 
-static void debug_shared(wchar_t msg_level, const wcstring &msg);
-
 #if defined(OS_IS_CYGWIN) || defined(WSL)
 // MS Windows tty devices do not currently have either a read or write timestamp. Those
 // respective fields of `struct stat` are always the current time. Which means we can't
@@ -224,17 +213,17 @@ bool is_windows_subsystem_for_linux() {
     return backtrace_text;
 }
 
-[[gnu::noinline]] void show_stackframe(const wchar_t msg_level, int frame_count, int skip_levels) {
+[[gnu::noinline]] void show_stackframe(int frame_count, int skip_levels) {
     if (frame_count < 1) return;
 
     wcstring_list_t bt = demangled_backtrace(frame_count, skip_levels + 2);
-    debug_shared(msg_level, L"Backtrace:\n" + join_strings(bt, L'\n') + L'\n');
+    FLOG(error, L"Backtrace:\n" + join_strings(bt, L'\n') + L'\n');
 }
 
 #else   // HAVE_BACKTRACE_SYMBOLS
 
-[[gnu::noinline]] void show_stackframe(const wchar_t msg_level, int, int) {
-    debug_shared(msg_level, L"Sorry, but your system does not support backtraces");
+[[gnu::noinline]] void show_stackframe(int, int) {
+    FLOGF(error, L"Sorry, but your system does not support backtraces");
 }
 #endif  // HAVE_BACKTRACE_SYMBOLS
 
@@ -250,7 +239,7 @@ static inline const char *align_start(const char *start, size_t len) {
     // How much do we have to add to start to make it 0 mod Align?
     // To compute 17 up-aligned by 8, compute its skew 17 % 8, yielding 1,
     // and then we will add 8 - 1. Of course if we align 16 with the same idea, we will
-    // add 8 instead of 0, so then mod the summand by Align again.
+    // add 8 instead of 0, so then mod the sum by Align again.
     // Note all of these mods are optimized to masks.
     uintptr_t add_which_aligns = Align - (startu % Align);
     add_which_aligns %= Align;
@@ -606,17 +595,6 @@ bool should_suppress_stderr_for_tests() {
     return program_name && !std::wcscmp(program_name, TESTS_PROGRAM_NAME);
 }
 
-static void debug_shared(const wchar_t level, const wcstring &msg) {
-    pid_t current_pid;
-    if (!is_forked_child()) {
-        std::fwprintf(stderr, L"<%lc> %ls: %ls\n", level, program_name, msg.c_str());
-    } else {
-        current_pid = getpid();
-        std::fwprintf(stderr, L"<%lc> %ls: %d: %ls\n", level, program_name, current_pid,
-                      msg.c_str());
-    }
-}
-
 // Careful to not negate LLONG_MIN.
 static unsigned long long absolute_value(long long x) {
     if (x >= 0) return static_cast<unsigned long long>(x);
@@ -661,6 +639,16 @@ void format_long_safe(wchar_t buff[64], long val) {
     }
 }
 
+void format_llong_safe(wchar_t buff[64], long long val) {
+    unsigned long long uval = absolute_value(val);
+    if (val >= 0) {
+        format_safe_impl(buff, 64, uval);
+    } else {
+        buff[0] = '-';
+        format_safe_impl(buff + 1, 63, uval);
+    }
+}
+
 void format_ullong_safe(wchar_t buff[64], unsigned long long val) {
     return format_safe_impl(buff, 64, val);
 }
@@ -681,12 +669,13 @@ void narrow_string_safe(char buff[64], const wchar_t *s) {
 
 wcstring reformat_for_screen(const wcstring &msg, const termsize_t &termsize) {
     wcstring buff;
-    int line_width = 0;
+
     int screen_width = termsize.width;
 
     if (screen_width) {
         const wchar_t *start = msg.c_str();
         const wchar_t *pos = start;
+        int line_width = 0;
         while (true) {
             int overflow = 0;
 
@@ -710,20 +699,18 @@ wcstring reformat_for_screen(const wcstring &msg, const termsize_t &termsize) {
                 pos = pos + 1;
             } else if (overflow) {
                 // In case of overflow, we print a newline, except if we already are at position 0.
-                wchar_t *token = wcsndup(start, pos - start);
+                wcstring token = msg.substr(start - msg.c_str(), pos - start);
                 if (line_width != 0) buff.push_back(L'\n');
-                buff.append(format_string(L"%ls-\n", token));
-                free(token);
+                buff.append(format_string(L"%ls-\n", token.c_str()));
                 line_width = 0;
             } else {
                 // Print the token.
-                wchar_t *token = wcsndup(start, pos - start);
+                wcstring token = msg.substr(start - msg.c_str(), pos - start);
                 if ((line_width + (line_width != 0 ? 1 : 0) + tok_width) > screen_width) {
                     buff.push_back(L'\n');
                     line_width = 0;
                 }
-                buff.append(format_string(L"%ls%ls", line_width ? L" " : L"", token));
-                free(token);
+                buff.append(format_string(L"%ls%ls", line_width ? L" " : L"", token.c_str()));
                 line_width += (line_width != 0 ? 1 : 0) + tok_width;
             }
 
@@ -883,11 +870,12 @@ wcstring escape_string_for_double_quotes(wcstring in) {
 static void escape_string_script(const wchar_t *orig_in, size_t in_len, wcstring &out,
                                  escape_flags_t flags) {
     const wchar_t *in = orig_in;
-    const bool escape_all = static_cast<bool>(flags & ESCAPE_ALL);
+    const bool escape_printables = !(flags & ESCAPE_NO_PRINTABLES);
     const bool no_quoted = static_cast<bool>(flags & ESCAPE_NO_QUOTED);
     const bool no_tilde = static_cast<bool>(flags & ESCAPE_NO_TILDE);
-    const bool no_caret = feature_test(features_t::stderr_nocaret);
     const bool no_qmark = feature_test(features_t::qmark_noglob);
+    const bool symbolic = static_cast<bool>(flags & ESCAPE_SYMBOLIC) && (MB_CUR_MAX > 1);
+    assert((!symbolic || !escape_printables) && "symbolic implies escape-no-printables");
 
     bool need_escape = false;
     bool need_complex_escape = false;
@@ -916,47 +904,57 @@ static void escape_string_script(const wchar_t *orig_in, size_t in_len, wcstring
             wchar_t c = *in;
             switch (c) {
                 case L'\t': {
-                    out += L'\\';
-                    out += L't';
+                    if (symbolic)
+                        out += L'␉';
+                    else
+                        out += L"\\t";
                     need_escape = need_complex_escape = true;
                     break;
                 }
                 case L'\n': {
-                    out += L'\\';
-                    out += L'n';
+                    if (symbolic)
+                        out += L'␤';
+                    else
+                        out += L"\\n";
                     need_escape = need_complex_escape = true;
                     break;
                 }
                 case L'\b': {
-                    out += L'\\';
-                    out += L'b';
+                    if (symbolic)
+                        out += L'␈';
+                    else
+                        out += L"\\b";
                     need_escape = need_complex_escape = true;
                     break;
                 }
                 case L'\r': {
-                    out += L'\\';
-                    out += L'r';
+                    if (symbolic)
+                        out += L'␍';
+                    else
+                        out += L"\\r";
                     need_escape = need_complex_escape = true;
                     break;
                 }
                 case L'\x1B': {
-                    out += L'\\';
-                    out += L'e';
+                    if (symbolic)
+                        out += L'␛';
+                    else
+                        out += L"\\e";
                     need_escape = need_complex_escape = true;
                     break;
                 }
                 case L'\x7F': {
-                    out += L'\\';
-                    out += L'x';
-                    out += L'7';
-                    out += L'f';
+                    if (symbolic)
+                        out += L'␡';
+                    else
+                        out += L"\\x7f";
                     need_escape = need_complex_escape = true;
                     break;
                 }
                 case L'\\':
                 case L'\'': {
                     need_escape = need_complex_escape = true;
-                    out += L'\\';
+                    if (escape_printables || (c == L'\\' && !symbolic)) out += L'\\';
                     out += *in;
                     break;
                 }
@@ -978,7 +976,6 @@ static void escape_string_script(const wchar_t *orig_in, size_t in_len, wcstring
                 case L'$':
                 case L' ':
                 case L'#':
-                case L'^':
                 case L'<':
                 case L'>':
                 case L'(':
@@ -994,24 +991,28 @@ static void escape_string_script(const wchar_t *orig_in, size_t in_len, wcstring
                 case L'"':
                 case L'%':
                 case L'~': {
-                    bool char_is_normal = (c == L'~' && no_tilde) || (c == L'^' && no_caret) ||
-                                          (c == L'?' && no_qmark);
+                    bool char_is_normal = (c == L'~' && no_tilde) || (c == L'?' && no_qmark);
                     if (!char_is_normal) {
                         need_escape = true;
-                        if (escape_all) out += L'\\';
+                        if (escape_printables) out += L'\\';
                     }
                     out += *in;
                     break;
                 }
 
                 default: {
-                    if (*in < 32) {
-                        if (*in < 27 && *in > 0) {
+                    if (*in >= 0 && *in < 32) {
+                        need_escape = need_complex_escape = true;
+
+                        if (symbolic) {
+                            out += L'\u2400' + *in;
+                            break;
+                        }
+
+                        if (*in < 27 && *in != 0) {
                             out += L'\\';
                             out += L'c';
                             out += L'a' + *in - 1;
-
-                            need_escape = need_complex_escape = true;
                             break;
                         }
 
@@ -1020,7 +1021,6 @@ static void escape_string_script(const wchar_t *orig_in, size_t in_len, wcstring
                         out += L'x';
                         out += ((*in > 15) ? L'1' : L'0');
                         out += tmp > 9 ? L'a' + (tmp - 10) : L'0' + tmp;
-                        need_escape = need_complex_escape = true;
                     } else {
                         out += *in;
                     }
@@ -1033,7 +1033,7 @@ static void escape_string_script(const wchar_t *orig_in, size_t in_len, wcstring
     }
 
     // Use quoted escaping if possible, since most people find it easier to read.
-    if (!no_quoted && need_escape && !need_complex_escape && escape_all) {
+    if (!no_quoted && need_escape && !need_complex_escape && escape_printables) {
         wchar_t single_quote = L'\'';
         out.clear();
         out.reserve(2 + in_len);
@@ -1044,8 +1044,7 @@ static void escape_string_script(const wchar_t *orig_in, size_t in_len, wcstring
 }
 
 /// Escapes a string for use in a regex string. Not safe for use with `eval` as only
-/// characters reserved by PCRE2 are escaped, i.e. it relies on fish's automatic escaping
-/// of subshell output in subsequent concatenation or for use as an argument.
+/// characters reserved by PCRE2 are escaped.
 /// \param in is the raw string to be searched for literally when substituted in a PCRE2 expression.
 static wcstring escape_string_pcre2(const wcstring &in) {
     wcstring out;
@@ -1072,7 +1071,7 @@ static wcstring escape_string_pcre2(const wcstring &in) {
             case L'-':
             case L']':
                 out.push_back('\\');
-                /* FALLTHROUGH */
+                __fallthrough__
             default:
                 out.push_back(c);
         }
@@ -1150,161 +1149,182 @@ maybe_t<size_t> read_unquoted_escape(const wchar_t *input, wcstring *result, boo
     bool errored = false;
     size_t in_pos = 1;  // in_pos always tracks the next character to read (and therefore the number
                         // of characters read so far)
-    const wchar_t c = input[in_pos++];
-    switch (c) {
-        // A null character after a backslash is an error.
-        case L'\0': {
-            // Adjust in_pos to only include the backslash.
-            assert(in_pos > 0);
-            in_pos--;
 
-            // It's an error, unless we're allowing incomplete escapes.
-            if (!allow_incomplete) errored = true;
-            break;
-        }
-        // Numeric escape sequences. No prefix means octal escape, otherwise hexadecimal.
-        case L'0':
-        case L'1':
-        case L'2':
-        case L'3':
-        case L'4':
-        case L'5':
-        case L'6':
-        case L'7':
-        case L'u':
-        case L'U':
-        case L'x':
-        case L'X': {
-            long long res = 0;
-            size_t chars = 2;
-            int base = 16;
-            bool byte_literal = false;
-            wchar_t max_val = ASCII_MAX;
+    // For multibyte \X sequences.
+    std::string byte_buff;
+    while (true) {
+        const wchar_t c = input[in_pos++];
+        switch (c) {
+                // A null character after a backslash is an error.
+            case L'\0': {
+                // Adjust in_pos to only include the backslash.
+                assert(in_pos > 0);
+                in_pos--;
 
-            switch (c) {
-                case L'u': {
-                    chars = 4;
-                    max_val = UCS2_MAX;
-                    break;
-                }
-                case L'U': {
-                    chars = 8;
-                    max_val = WCHAR_MAX;
-
-                    // Don't exceed the largest Unicode code point - see #1107.
-                    if (0x10FFFF < max_val) max_val = static_cast<wchar_t>(0x10FFFF);
-                    break;
-                }
-                case L'x': {
-                    chars = 2;
-                    max_val = ASCII_MAX;
-                    break;
-                }
-                case L'X': {
-                    byte_literal = true;
-                    max_val = BYTE_MAX;
-                    break;
-                }
-                default: {
-                    base = 8;
-                    chars = 3;
-                    // Note that in_pos currently is just after the first post-backslash character;
-                    // we want to start our escape from there.
-                    assert(in_pos > 0);
-                    in_pos--;
-                    break;
-                }
+                // It's an error, unless we're allowing incomplete escapes.
+                if (!allow_incomplete) errored = true;
+                break;
             }
+                // Numeric escape sequences. No prefix means octal escape, otherwise hexadecimal.
+            case L'0':
+            case L'1':
+            case L'2':
+            case L'3':
+            case L'4':
+            case L'5':
+            case L'6':
+            case L'7':
+            case L'u':
+            case L'U':
+            case L'x':
+            case L'X': {
+                long long res = 0;
+                size_t chars = 2;
+                int base = 16;
+                bool byte_literal = false;
+                wchar_t max_val = ASCII_MAX;
 
-            for (size_t i = 0; i < chars; i++) {
-                long d = convert_digit(input[in_pos], base);
-                if (d < 0) {
-                    // If we have no digit, this is a tokenizer error.
-                    if (i == 0) errored = true;
-                    break;
+                switch (c) {
+                    case L'u': {
+                        chars = 4;
+                        max_val = UCS2_MAX;
+                        break;
+                    }
+                    case L'U': {
+                        chars = 8;
+                        max_val = WCHAR_MAX;
+
+                        // Don't exceed the largest Unicode code point - see #1107.
+                        if (0x10FFFF < max_val) max_val = static_cast<wchar_t>(0x10FFFF);
+                        break;
+                    }
+                    case L'x':
+                    case L'X': {
+                        byte_literal = true;
+                        max_val = BYTE_MAX;
+                        break;
+                    }
+                    default: {
+                        base = 8;
+                        chars = 3;
+                        // Note that in_pos currently is just after the first post-backslash
+                        // character; we want to start our escape from there.
+                        assert(in_pos > 0);
+                        in_pos--;
+                        break;
+                    }
                 }
 
-                res = (res * base) + d;
-                in_pos++;
-            }
+                for (size_t i = 0; i < chars; i++) {
+                    long d = convert_digit(input[in_pos], base);
+                    if (d < 0) {
+                        // If we have no digit, this is a tokenizer error.
+                        if (i == 0) errored = true;
+                        break;
+                    }
 
-            if (!errored && res <= max_val) {
-                result_char_or_none =
-                    static_cast<wchar_t>((byte_literal ? ENCODE_DIRECT_BASE : 0) + res);
-            } else {
-                errored = true;
-            }
+                    res = (res * base) + d;
+                    in_pos++;
+                }
 
-            break;
-        }
-        // \a means bell (alert).
-        case L'a': {
-            result_char_or_none = L'\a';
-            break;
-        }
-        // \b means backspace.
-        case L'b': {
-            result_char_or_none = L'\b';
-            break;
-        }
-        // \cX means control sequence X.
-        case L'c': {
-            const wchar_t sequence_char = input[in_pos++];
-            if (sequence_char >= L'a' && sequence_char <= (L'a' + 32)) {
-                result_char_or_none = sequence_char - L'a' + 1;
-            } else if (sequence_char >= L'A' && sequence_char <= (L'A' + 32)) {
-                result_char_or_none = sequence_char - L'A' + 1;
-            } else {
-                errored = true;
+                if (!errored && res <= max_val) {
+                    if (byte_literal) {
+                        // Multibyte encodings necessitate that we keep adjacent byte escapes.
+                        // - `\Xc3\Xb6` is "ö", but only together.
+                        // (this assumes a valid codepoint can't consist of multiple bytes
+                        // that are valid on their own, which is true for UTF-8)
+                        byte_buff.push_back(static_cast<char>(res));
+                        result_char_or_none = none();
+                        if (input[in_pos] == L'\\'
+                            && (input[in_pos + 1] == L'X' || input[in_pos + 1] == L'x')) {
+                            in_pos++;
+                            continue;
+                        }
+                    } else {
+                        result_char_or_none = static_cast<wchar_t>(res);
+                    }
+                } else {
+                    errored = true;
+                }
+
+                break;
+            }
+                // \a means bell (alert).
+            case L'a': {
+                result_char_or_none = L'\a';
+                break;
+            }
+                // \b means backspace.
+            case L'b': {
+                result_char_or_none = L'\b';
+                break;
+            }
+                // \cX means control sequence X.
+            case L'c': {
+                const wchar_t sequence_char = input[in_pos++];
+                if (sequence_char >= L'a' && sequence_char <= (L'a' + 32)) {
+                    result_char_or_none = sequence_char - L'a' + 1;
+                } else if (sequence_char >= L'A' && sequence_char <= (L'A' + 32)) {
+                    result_char_or_none = sequence_char - L'A' + 1;
+                } else {
+                    errored = true;
+                }
+                break;
+            }
+                // \x1B means escape.
+            case L'e': {
+                result_char_or_none = L'\x1B';
+                break;
+            }
+                // \f means form feed.
+            case L'f': {
+                result_char_or_none = L'\f';
+                break;
+            }
+                // \n means newline.
+            case L'n': {
+                result_char_or_none = L'\n';
+                break;
+            }
+                // \r means carriage return.
+            case L'r': {
+                result_char_or_none = L'\r';
+                break;
+            }
+                // \t means tab.
+            case L't': {
+                result_char_or_none = L'\t';
+                break;
+            }
+                // \v means vertical tab.
+            case L'v': {
+                result_char_or_none = L'\v';
+                break;
+            }
+                // If a backslash is followed by an actual newline, swallow them both.
+            case L'\n': {
+                result_char_or_none = none();
+                break;
+            }
+            default: {
+                if (unescape_special) result->push_back(INTERNAL_SEPARATOR);
+                result_char_or_none = c;
+                break;
             }
-            break;
         }
-        // \x1B means escape.
-        case L'e': {
-            result_char_or_none = L'\x1B';
-            break;
-        }
-        // \f means form feed.
-        case L'f': {
-            result_char_or_none = L'\f';
-            break;
-        }
-        // \n means newline.
-        case L'n': {
-            result_char_or_none = L'\n';
-            break;
-        }
-        // \r means carriage return.
-        case L'r': {
-            result_char_or_none = L'\r';
-            break;
-        }
-        // \t means tab.
-        case L't': {
-            result_char_or_none = L'\t';
-            break;
-        }
-        // \v means vertical tab.
-        case L'v': {
-            result_char_or_none = L'\v';
-            break;
-        }
-        // If a backslash is followed by an actual newline, swallow them both.
-        case L'\n': {
-            result_char_or_none = none();
-            break;
-        }
-        default: {
-            if (unescape_special) result->push_back(INTERNAL_SEPARATOR);
-            result_char_or_none = c;
-            break;
+
+        if (errored) return none();
+
+        if (!byte_buff.empty()) {
+            result->append(str2wcstring(byte_buff));
         }
+
+        break;
     }
 
-    if (!errored && result_char_or_none.has_value()) {
+    if (result_char_or_none.has_value()) {
         result->push_back(*result_char_or_none);
     }
-    if (errored) return none();
 
     return in_pos;
 }
@@ -1347,7 +1367,7 @@ static bool unescape_string_internal(const wchar_t *const input, const size_t in
                         // appending INTERNAL_SEPARATORs, so we have to handle them specially.
                         auto escape_chars = read_unquoted_escape(
                             input + input_position, &result, allow_incomplete, unescape_special);
-                        if (!escape_chars) {
+                        if (!escape_chars.has_value()) {
                             // A none() return indicates an error.
                             errored = true;
                         } else {
@@ -1429,8 +1449,9 @@ static bool unescape_string_internal(const wchar_t *const input, const size_t in
                         brace_count--;
                         to_append_or_none = BRACE_END;
                         if (!braces.empty()) {
-                            // If we didn't have a var or separator since the last '{',
-                            // put the literal back.
+                            // HACK: To reduce accidental use of brace expansion, treat a brace
+                            // with zero or one items as literal input. See #4632. (The hack is
+                            // doing it here and like this.)
                             if (vars_or_seps.empty() || vars_or_seps.back() < braces.back()) {
                                 result[braces.back()] = L'{';
                                 // We also need to turn all spaces back.
@@ -1624,11 +1645,6 @@ bool unescape_string(const wcstring &input, wcstring *output, unescape_flags_t e
     return unescape_string(input.c_str(), input.size(), output, escape_special, style);
 }
 
-[[gnu::noinline]] void bugreport() {
-    FLOG(error, _(L"This is a bug. Break on 'bugreport' to debug."));
-    FLOG(error, _(L"If you can reproduce it, please report: "), PACKAGE_BUGREPORT, L'.');
-}
-
 wcstring format_size(long long sz) {
     wcstring result;
     const wchar_t *sz_name[] = {L"kB", L"MB", L"GB", L"TB", L"PB", L"EB", L"ZB", L"YB", nullptr};
@@ -1689,7 +1705,7 @@ void format_size_safe(char buff[128], unsigned long long sz) {
     size_t idx = 0;
     const char *const sz_name[] = {"kB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB", nullptr};
     if (sz < 1) {
-        strncpy(buff, "empty", buff_size);
+        strcpy(buff, "empty");
     } else if (sz < 1024) {
         append_ull(buff, sz, &idx, max_len);
         append_str(buff, "B", &idx, max_len);
@@ -1752,10 +1768,7 @@ void setup_fork_guards() {
                    [] { pthread_atfork(nullptr, nullptr, [] { is_forked_proc = true; }); });
 }
 
-void save_term_foreground_process_group() {
-    ASSERT_IS_MAIN_THREAD();
-    initial_fg_process_group = tcgetpgrp(STDIN_FILENO);
-}
+void save_term_foreground_process_group() { initial_fg_process_group = tcgetpgrp(STDIN_FILENO); }
 
 void restore_term_foreground_process_group_for_exit() {
     // We wish to restore the tty to the initial owner. There's two ways this can go wrong:
@@ -1808,7 +1821,7 @@ void assert_is_locked(std::mutex &mutex, const char *who, const char *caller) {
     }
 }
 
-/// Test if the specified character is in a range that fish uses interally to store special tokens.
+/// Test if the specified character is in a range that fish uses internally to store special tokens.
 ///
 /// NOTE: This is used when tokenizing the input. It is also used when reading input, before
 /// tokenization, to replace such chars with REPLACEMENT_WCHAR if they're not part of a quoted
@@ -1843,7 +1856,7 @@ void redirect_tty_output() {
     } else {
         FLOGF(error, L"%s:%zu: failed assertion: %s", file, line, msg);
     }
-    show_stackframe(L'E', 99, 1);
+    show_stackframe(99, 1);
     abort();
 }
 
@@ -1884,14 +1897,16 @@ std::string get_executable_path(const char *argv0) {
     // https://opensource.apple.com/source/adv_cmds/adv_cmds-163/ps/print.c
     uint32_t buffSize = sizeof buff;
     if (_NSGetExecutablePath(buff, &buffSize) == 0) return std::string(buff);
-#elif defined(__BSD__) && defined(KERN_PROC_PATHNAME) && !defined(__NetBSD__)
+#elif defined(__BSD__) && defined(KERN_PROC_PATHNAME)
     // BSDs do not have /proc by default, (although it can be mounted as procfs via the Linux
     // compatibility layer). We can use sysctl instead: per sysctl(3), passing in a process ID of -1
     // returns the value for the current process.
-    //
-    // (this is broken on NetBSD, while /proc works, so we use that)
     size_t buff_size = sizeof buff;
+#if defined(__NetBSD__)
+    int name[] = {CTL_KERN, KERN_PROC_ARGS, getpid(), KERN_PROC_PATHNAME};
+#else
     int name[] = {CTL_KERN, KERN_PROC, KERN_PROC_PATHNAME, -1};
+#endif
     int result = sysctl(name, sizeof(name) / sizeof(int), buff, &buff_size, nullptr, 0);
     if (result != 0) {
         wperror(L"sysctl KERN_PROC_PATHNAME");
@@ -1910,7 +1925,18 @@ std::string get_executable_path(const char *argv0) {
     }
     if (len > 0) {
         buff[len] = '\0';
-        return std::string(buff);
+        // When /proc/self/exe points to a file that was deleted (or overwritten on update!)
+        // then linux adds a " (deleted)" suffix.
+        // If that's not a valid path, let's remove that awkward suffix.
+        std::string buffstr{buff};
+        if (access(buff, F_OK)) {
+            auto dellen = const_strlen(" (deleted)");
+            if (buffstr.size() > dellen &&
+                buffstr.compare(buffstr.size() - dellen, dellen, " (deleted)") == 0) {
+                buffstr = buffstr.substr(0, buffstr.size() - dellen);
+            }
+        }
+        return buffstr;
     }
 #endif
 
@@ -1951,14 +1977,14 @@ std::string get_path_to_tmp_dir() {
 // bullet-proof and that's OK.
 bool is_console_session() {
     static const bool console_session = [] {
-        ASSERT_IS_MAIN_THREAD();
-
-        const char *tty_name = ttyname(0);
+        char tty_name[PATH_MAX];
+        if (ttyname_r(STDIN_FILENO, tty_name, sizeof tty_name) != 0) {
+            return false;
+        }
         constexpr auto len = const_strlen("/dev/tty");
         const char *TERM = getenv("TERM");
         return
             // Test that the tty matches /dev/(console|dcons|tty[uv\d])
-            tty_name &&
             ((strncmp(tty_name, "/dev/tty", len) == 0 &&
               (tty_name[len] == 'u' || tty_name[len] == 'v' || isdigit(tty_name[len]))) ||
              strcmp(tty_name, "/dev/dcons") == 0 || strcmp(tty_name, "/dev/console") == 0)
