@@ -10,14 +10,19 @@
 #ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>  // IWYU pragma: keep
 #endif
+#include <sys/types.h>
+#include <termios.h>
 
 #include <algorithm>
-#include <atomic>
-#include <cmath>
+#include <cstdint>
+#include <cwchar>
 #include <functional>
+#include <initializer_list>
+#include <iterator>
 #include <memory>
 #include <mutex>
 #include <string>
+#include <utility>
 #include <vector>
 
 #include "fallback.h"  // IWYU pragma: keep
@@ -68,18 +73,18 @@ struct termsize_t;
 // Unicode BOM value.
 #define UTF8_BOM_WCHAR 0xFEFFu
 
-// Use Unicode "noncharacters" for internal characters as much as we can. This
+// Use Unicode "non-characters" for internal characters as much as we can. This
 // gives us 32 "characters" for internal use that we can guarantee should not
 // appear in our input stream. See http://www.unicode.org/faq/private_use.html.
 #define RESERVED_CHAR_BASE static_cast<wchar_t>(0xFDD0)
 #define RESERVED_CHAR_END static_cast<wchar_t>(0xFDF0)
-// Split the available noncharacter values into two ranges to ensure there are
+// Split the available non-character values into two ranges to ensure there are
 // no conflicts among the places we use these special characters.
 #define EXPAND_RESERVED_BASE RESERVED_CHAR_BASE
 #define EXPAND_RESERVED_END (EXPAND_RESERVED_BASE + 16)
 #define WILDCARD_RESERVED_BASE EXPAND_RESERVED_END
 #define WILDCARD_RESERVED_END (WILDCARD_RESERVED_BASE + 16)
-// Make sure the ranges defined above don't exceed the range for noncharacters.
+// Make sure the ranges defined above don't exceed the range for non-characters.
 // This is to make sure we didn't do something stupid in subdividing the
 // Unicode range for our needs.
 //#if WILDCARD_RESERVED_END > RESERVED_CHAR_END
@@ -138,16 +143,19 @@ enum {
 };
 typedef unsigned int unescape_flags_t;
 
-// Flags for the escape_string() and escape_string() functions. These are only applicable when the
-// escape style is "script" (i.e., STRING_STYLE_SCRIPT).
+// Flags for the escape_string() function. These are only applicable when the escape style is
+// "script" (i.e., STRING_STYLE_SCRIPT).
 enum {
-    /// Escape all characters, including magic characters like the semicolon.
-    ESCAPE_ALL = 1 << 0,
+    /// Do not escape special fish syntax characters like the semicolon. Only escape non-printable
+    /// characters and backslashes.
+    ESCAPE_NO_PRINTABLES = 1 << 0,
     /// Do not try to use 'simplified' quoted escapes, and do not use empty quotes as the empty
     /// string.
     ESCAPE_NO_QUOTED = 1 << 1,
     /// Do not escape tildes.
-    ESCAPE_NO_TILDE = 1 << 2
+    ESCAPE_NO_TILDE = 1 << 2,
+    /// Replace non-printable control characters with Unicode symbols.
+    ESCAPE_SYMBOLIC = 1 << 3
 };
 typedef unsigned int escape_flags_t;
 
@@ -196,7 +204,7 @@ extern const wcstring g_empty_string;
 #define FATAL_EXIT()                                \
     do {                                            \
         char exit_read_buff;                        \
-        show_stackframe(L'E');                      \
+        show_stackframe();                          \
         ignore_result(read(0, &exit_read_buff, 1)); \
         exit_without_destructors(1);                \
     } while (0)
@@ -207,8 +215,8 @@ extern const wcstring g_empty_string;
 /// stdio functions and should be writing the message to stderr rather than stdout. Second, if
 /// possible it is useful to provide additional context such as a stack backtrace.
 #undef assert
-#define assert(e) (e) ? ((void)0) : __fish_assert(#e, __FILE__, __LINE__, 0)
-#define assert_with_errno(e) (e) ? ((void)0) : __fish_assert(#e, __FILE__, __LINE__, errno)
+#define assert(e) likely(e) ? ((void)0) : __fish_assert(#e, __FILE__, __LINE__, 0)
+#define assert_with_errno(e) likely(e) ? ((void)0) : __fish_assert(#e, __FILE__, __LINE__, errno)
 #define DIE(msg) __fish_assert(msg, __FILE__, __LINE__, 0)
 #define DIE_WITH_ERRNO(msg) __fish_assert(msg, __FILE__, __LINE__, errno)
 /// This macro is meant to be used with functions that return zero on success otherwise return an
@@ -254,6 +262,11 @@ bool contains(const Col &col, const T2 &val) {
     return std::find(std::begin(col), std::end(col), val) != std::end(col);
 }
 
+template <typename T1, typename T2>
+bool contains(std::initializer_list<T1> col, const T2 &val) {
+    return std::find(std::begin(col), std::end(col), val) != std::end(col);
+}
+
 /// Append a vector \p donator to the vector \p receiver.
 template <typename T>
 void vec_append(std::vector<T> &receiver, std::vector<T> &&donator) {
@@ -265,18 +278,12 @@ void vec_append(std::vector<T> &receiver, std::vector<T> &&donator) {
     }
 }
 
-/// Move an object into a shared_ptr.
-template <typename T>
-std::shared_ptr<T> move_to_sharedptr(T &&v) {
-    return std::make_shared<T>(std::move(v));
-}
-
 /// A function type to check for cancellation.
 /// \return true if execution should cancel.
 using cancel_checker_t = std::function<bool()>;
 
 /// Print a stack trace to stderr.
-void show_stackframe(const wchar_t msg_level, int frame_count = 100, int skip_levels = 0);
+void show_stackframe(int frame_count = 100, int skip_levels = 0);
 
 /// Returns a  wide character string equivalent of the specified multibyte character string.
 ///
@@ -303,8 +310,8 @@ void wcs2string_appending(const wchar_t *in, size_t len, std::string *receiver);
 bool should_suppress_stderr_for_tests();
 
 /// Branch prediction hints. Idea borrowed from Linux kernel. Just used for asserts.
-#define likely(x) __builtin_expect(!!(x), 1)
-#define unlikely(x) __builtin_expect(!!(x), 0)
+#define likely(x) __builtin_expect(bool(x), 1)
+#define unlikely(x) __builtin_expect(bool(x), 0)
 
 void assert_is_main_thread(const char *who);
 #define ASSERT_IS_MAIN_THREAD_TRAMPOLINE(x) assert_is_main_thread(x)
@@ -328,11 +335,15 @@ void format_size_safe(char buff[128], unsigned long long sz);
 /// Writes out a long safely.
 void format_long_safe(char buff[64], long val);
 void format_long_safe(wchar_t buff[64], long val);
+void format_llong_safe(wchar_t buff[64], long long val);
 void format_ullong_safe(wchar_t buff[64], unsigned long long val);
 
-/// "Narrows" a wide character string. This just grabs any ASCII characters and trunactes.
+/// "Narrows" a wide character string. This just grabs any ASCII characters and truncates.
 void narrow_string_safe(char buff[64], const wchar_t *s);
 
+/// Stored in blocks to reference the file which created the block.
+using filename_ref_t = std::shared_ptr<const wcstring>;
+
 using scoped_lock = std::lock_guard<std::mutex>;
 
 // An object wrapping a scoped lock and a value
@@ -477,9 +488,9 @@ ssize_t read_loop(int fd, void *buff, size_t count);
 /// \param in The string to be escaped
 /// \param flags Flags to control the escaping
 /// \return The escaped string
-wcstring escape_string(const wchar_t *in, escape_flags_t flags,
+wcstring escape_string(const wchar_t *in, escape_flags_t flags = 0,
                        escape_string_style_t style = STRING_STYLE_SCRIPT);
-wcstring escape_string(const wcstring &in, escape_flags_t flags,
+wcstring escape_string(const wcstring &in, escape_flags_t flags = 0,
                        escape_string_style_t style = STRING_STYLE_SCRIPT);
 
 /// Escape a string so that it may be inserted into a double-quoted string.
@@ -550,7 +561,7 @@ void assert_is_not_forked_child(const char *who);
 /// See https://github.com/Microsoft/WSL/issues/423 and Microsoft/WSL#2997
 bool is_windows_subsystem_for_linux();
 
-/// Detect if we are running under Cygwin or Cgywin64
+/// Detect if we are running under Cygwin or Cygwin64
 constexpr bool is_cygwin() {
 #ifdef __CYGWIN__
     return true;
