@@ -12,13 +12,13 @@
 #include <unistd.h>
 
 #include <algorithm>
-#include <iterator>
+#include <map>
 #include <mutex>
 #include <set>
 #include <utility>
 #include <vector>
 
-#include "builtins/bind.h"
+#include "abbrs.h"
 #include "common.h"
 #include "env_dispatch.h"
 #include "env_universal_common.h"
@@ -30,6 +30,7 @@
 #include "history.h"
 #include "input.h"
 #include "kill.h"
+#include "null_terminated_array.h"
 #include "path.h"
 #include "proc.h"
 #include "reader.h"
@@ -65,11 +66,12 @@ static acquired_lock<env_universal_t> uvars() {
     return s_universal_variables->acquire();
 }
 
+/// Set when a universal variable has been modified but not yet been written to disk via sync().
+static relaxed_atomic_bool_t s_uvars_locally_modified{false};
+
 /// Whether we were launched with no_config; in this case setting a uvar instead sets a global.
 static relaxed_atomic_bool_t s_uvar_scope_is_global{false};
 
-bool env_universal_barrier() { return env_stack_t::principal().universal_barrier(); }
-
 namespace {
 struct electric_var_t {
     enum {
@@ -193,24 +195,61 @@ maybe_t<env_var_t> null_environment_t::get(const wcstring &key, env_mode_flags_t
     UNUSED(mode);
     return none();
 }
-wcstring_list_t null_environment_t::get_names(int flags) const {
+wcstring_list_t null_environment_t::get_names(env_mode_flags_t flags) const {
     UNUSED(flags);
     return {};
 }
 
-/// Set up the USER variable.
-static void setup_user(bool force) {
-    auto &vars = env_stack_t::globals();
-    if (force || vars.get(L"USER").missing_or_empty()) {
-        struct passwd userinfo;
-        struct passwd *result;
-        char buf[8192];
-        int retval = getpwuid_r(getuid(), &userinfo, buf, sizeof(buf), &result);
+/// Set up the USER and HOME variable.
+static void setup_user(env_stack_t &vars) {
+    auto uid = geteuid();
+    auto user_var = vars.get(L"USER");
+    struct passwd userinfo;
+    struct passwd *result;
+    char buf[8192];
+
+    // If we have a $USER, we try to get the passwd entry for the name.
+    // If that has the same UID that we use, we assume the data is correct.
+    if (!user_var.missing_or_empty()) {
+        std::string unam_narrow = wcs2string(user_var->as_string());
+        int retval = getpwnam_r(unam_narrow.c_str(), &userinfo, buf, sizeof(buf), &result);
         if (!retval && result) {
-            const wcstring uname = str2wcstring(userinfo.pw_name);
-            vars.set_one(L"USER", ENV_GLOBAL | ENV_EXPORT, uname);
+            if (result->pw_uid == uid) {
+                // The uid matches but we still might need to set $HOME.
+                if (vars.get(L"HOME").missing_or_empty()) {
+                    if (userinfo.pw_dir) {
+                        vars.set_one(L"HOME", ENV_GLOBAL | ENV_EXPORT,
+                                     str2wcstring(userinfo.pw_dir));
+                    } else {
+                        vars.set_empty(L"HOME", ENV_GLOBAL | ENV_EXPORT);
+                    }
+                }
+                return;
+            }
         }
     }
+
+    // Either we didn't have a $USER or it had a different uid.
+    // We need to get the data *again* via the uid.
+    int retval = getpwuid_r(uid, &userinfo, buf, sizeof(buf), &result);
+    if (!retval && result) {
+        const wcstring uname = str2wcstring(userinfo.pw_name);
+        vars.set_one(L"USER", ENV_GLOBAL | ENV_EXPORT, uname);
+        // Only change $HOME if it's empty, so we allow e.g. `HOME=(mktemp -d)`.
+        // This is okay with common `su` and `sudo` because they set $HOME.
+        if (vars.get(L"HOME").missing_or_empty()) {
+            if (userinfo.pw_dir) {
+                vars.set_one(L"HOME", ENV_GLOBAL | ENV_EXPORT, str2wcstring(userinfo.pw_dir));
+            } else {
+                // We cannot get $HOME. This triggers warnings for history and config.fish already,
+                // so it isn't necessary to warn here as well.
+                vars.set_empty(L"HOME", ENV_GLOBAL | ENV_EXPORT);
+            }
+        }
+    } else if (vars.get(L"HOME").missing_or_empty()) {
+        // If $USER is empty as well (which we tried to set above), we can't get $HOME.
+        vars.set_empty(L"HOME", ENV_GLOBAL | ENV_EXPORT);
+    }
 }
 
 /// Various things we need to initialize at run-time that don't really fit any of the other init
@@ -246,6 +285,10 @@ static void setup_path() {
     }
 }
 
+static std::map<wcstring, wcstring> inheriteds;
+
+const std::map<wcstring, wcstring> &env_get_inherited() { return inheriteds; }
+
 void env_init(const struct config_paths_t *paths, bool do_uvars, bool default_paths) {
     env_stack_t &vars = env_stack_t::principal();
     // Import environment variables. Walk backwards so that the first one out of any duplicates wins
@@ -262,9 +305,11 @@ void env_init(const struct config_paths_t *paths, bool do_uvars, bool default_pa
             if (!electric_var_t::for_name(key_and_val)) {
                 vars.set_empty(key_and_val, ENV_EXPORT | ENV_GLOBAL);
             }
+            inheriteds[key] = L"";
         } else {
             key.assign(key_and_val, 0, eql);
             val.assign(key_and_val, eql + 1, wcstring::npos);
+            inheriteds[key] = val;
             if (!electric_var_t::for_name(key)) {
                 // fish_user_paths should not be exported; attempting to re-import it from
                 // a value we previously (due to user error) exported will cause impossibly
@@ -289,51 +334,10 @@ void env_init(const struct config_paths_t *paths, bool do_uvars, bool default_pa
         }
     }
 
-    // Some `su`s keep $USER when changing to root.
-    // This leads to issues later on (and e.g. in prompts),
-    // so we work around it by resetting $USER.
-    // TODO: Figure out if that su actually checks if username == "root"(as the man page says) or
-    // UID == 0.
-    uid_t uid = getuid();
-    setup_user(uid == 0);
-
-    // Set up the HOME variable.
-    // Unlike $USER, it doesn't seem that `su`s pass this along
-    // if the target user is root, unless "--preserve-environment" is used.
-    // Since that is an explicit choice, we should allow it to enable e.g.
-    //     env HOME=(mktemp -d) su --preserve-environment fish
-    //
-    // Note: This needs to be *before* path_get_*, because that uses $HOME!
-    if (vars.get(L"HOME").missing_or_empty()) {
-        auto user_var = vars.get(L"USER");
-        if (!user_var.missing_or_empty()) {
-            std::string unam_narrow = wcs2string(user_var->as_string());
-            struct passwd userinfo;
-            struct passwd *result;
-            char buf[8192];
-            int retval = getpwnam_r(unam_narrow.c_str(), &userinfo, buf, sizeof(buf), &result);
-            if (retval || !result) {
-                // Maybe USER is set but it's bogus. Reset USER from the db and try again.
-                setup_user(true);
-                user_var = vars.get(L"USER");
-                if (!user_var.missing_or_empty()) {
-                    unam_narrow = wcs2string(user_var->as_string());
-                    retval = getpwnam_r(unam_narrow.c_str(), &userinfo, buf, sizeof(buf), &result);
-                }
-            }
-            if (!retval && result && userinfo.pw_dir) {
-                const wcstring dir = str2wcstring(userinfo.pw_dir);
-                vars.set_one(L"HOME", ENV_GLOBAL | ENV_EXPORT, dir);
-            } else {
-                // We cannot get $HOME. This triggers warnings for history and config.fish already,
-                // so it isn't necessary to warn here as well.
-                vars.set_empty(L"HOME", ENV_GLOBAL | ENV_EXPORT);
-            }
-        } else {
-            // If $USER is empty as well (which we tried to set above), we can't get $HOME.
-            vars.set_empty(L"HOME", ENV_GLOBAL | ENV_EXPORT);
-        }
-    }
+    // Set $USER, $HOME and $EUID
+    // This involves going to passwd and stuff.
+    vars.set_one(L"EUID", ENV_GLOBAL, to_string(static_cast<unsigned long long>(geteuid())));
+    setup_user(vars);
 
     wcstring user_config_dir;
     path_get_config(user_config_dir);
@@ -349,6 +353,11 @@ void env_init(const struct config_paths_t *paths, bool do_uvars, bool default_pa
     // Set up $IFS - this used to be in share/config.fish, but really breaks if it isn't done.
     vars.set_one(L"IFS", ENV_GLOBAL, L"\n \t");
 
+    // Ensure this var is present even before an interactive command is run so that if it is used
+    // in a function like `fish_prompt` or `fish_right_prompt` it is defined at the time the first
+    // prompt is written.
+    vars.set_one(L"CMD_DURATION", ENV_UNEXPORT, L"0");
+
     // Set up the version variable.
     wcstring version = str2wcstring(get_fish_version());
     vars.set_one(L"version", ENV_GLOBAL, version);
@@ -427,7 +436,9 @@ void env_init(const struct config_paths_t *paths, bool do_uvars, bool default_pa
         // Set up universal variables using the default path.
         callback_data_list_t callbacks;
         uvars()->initialize(callbacks);
-        env_universal_callbacks(&vars, callbacks);
+        for (const callback_data_t &cb : callbacks) {
+            env_dispatch_var_change(cb.key, vars);
+        }
 
         // Do not import variables that have the same name and value as
         // an exported universal variable. See issues #5258 and #5348.
@@ -442,6 +453,10 @@ void env_init(const struct config_paths_t *paths, bool do_uvars, bool default_pa
                 vars.globals().remove(name, ENV_GLOBAL | ENV_EXPORT);
             }
         }
+
+        // Import any abbreviations from uvars.
+        // Note we do not dynamically react to changes.
+        abbrs_get_set()->import_from_uvars(table);
     }
 }
 
@@ -565,7 +580,7 @@ class env_scoped_impl_t : public environment_t, noncopyable_t {
     }
 
     maybe_t<env_var_t> get(const wcstring &key, env_mode_flags_t mode = ENV_DEFAULT) const override;
-    wcstring_list_t get_names(int flags) const override;
+    wcstring_list_t get_names(env_mode_flags_t flags) const override;
 
     perproc_data_t &perproc_data() { return perproc_data_; }
     const perproc_data_t &perproc_data() const { return perproc_data_; }
@@ -767,10 +782,10 @@ maybe_t<env_var_t> env_scoped_impl_t::try_get_computed(const wcstring &key) cons
 
 maybe_t<env_var_t> env_scoped_impl_t::try_get_local(const wcstring &key) const {
     maybe_t<env_var_t> entry;
-    for (auto cur = locals_; cur; cur=cur->next) {
+    for (auto cur = locals_; cur; cur = cur->next) {
         if ((entry = cur->find_entry(key))) break;
     }
-    return entry; // this is either the entry or none() from find_entry
+    return entry;  // this is either the entry or none() from find_entry
 }
 
 maybe_t<env_var_t> env_scoped_impl_t::try_get_function(const wcstring &key) const {
@@ -783,10 +798,10 @@ maybe_t<env_var_t> env_scoped_impl_t::try_get_function(const wcstring &key) cons
         // topmost local scope.
         if (node->new_scope) break;
     }
-    for (auto cur = node; cur; cur=cur->next) {
+    for (auto cur = node; cur; cur = cur->next) {
         if ((entry = cur->find_entry(key))) break;
     }
-    return entry; // this is either the entry or none() from find_entry
+    return entry;  // this is either the entry or none() from find_entry
 }
 
 maybe_t<env_var_t> env_scoped_impl_t::try_get_global(const wcstring &key) const {
@@ -829,7 +844,7 @@ maybe_t<env_var_t> env_scoped_impl_t::get(const wcstring &key, env_mode_flags_t
     return result;
 }
 
-wcstring_list_t env_scoped_impl_t::get_names(int flags) const {
+wcstring_list_t env_scoped_impl_t::get_names(env_mode_flags_t flags) const {
     const query_t query(flags);
     std::set<wcstring> names;
 
@@ -1127,7 +1142,6 @@ maybe_t<int> env_stack_impl_t::try_set_electric(const wcstring &key, const query
 /// Set a universal variable, inheriting as applicable from the given old variable.
 void env_stack_impl_t::set_universal(const wcstring &key, wcstring_list_t val,
                                      const query_t &query) {
-    ASSERT_IS_MAIN_THREAD();
     auto oldvar = uvars()->get(key);
     // Resolve whether or not to export.
     bool exports = false;
@@ -1170,7 +1184,8 @@ mod_result_t env_stack_impl_t::set(const wcstring &key, env_mode_flags_t mode,
                                    wcstring_list_t val) {
     const query_t query(mode);
     // Handle electric and read-only variables.
-    if (auto ret = try_set_electric(key, query, val)) {
+    auto ret = try_set_electric(key, query, val);
+    if (ret.has_value()) {
         return mod_result_t{*ret};
     }
 
@@ -1286,23 +1301,25 @@ mod_result_t env_stack_impl_t::remove(const wcstring &key, int mode) {
     return result;
 }
 
-bool env_stack_t::universal_barrier() {
-    // Only perform universal barriers for the principal env stack.
-    // This means that changes from other fish processes will only be visible when the "main thread
-    // runs."
-    if (!is_principal()) return false;
-
-    ASSERT_IS_MAIN_THREAD();
-    if (s_uvar_scope_is_global) return false;
+std::vector<event_t> env_stack_t::universal_sync(bool always) {
+    if (s_uvar_scope_is_global) return {};
+    if (!always && !s_uvars_locally_modified) return {};
+    s_uvars_locally_modified = false;
 
     callback_data_list_t callbacks;
     bool changed = uvars()->sync(callbacks);
     if (changed) {
         universal_notifier_t::default_notifier().post_notification();
     }
-
-    env_universal_callbacks(this, callbacks);
-    return changed || !callbacks.empty();
+    // React internally to changes to special variables like LANG, and populate on-variable events.
+    std::vector<event_t> result;
+    for (const callback_data_t &cb : callbacks) {
+        env_dispatch_var_change(cb.key, *this);
+        event_t evt =
+            cb.is_erase() ? event_t::variable_erase(cb.key) : event_t::variable_set(cb.key);
+        result.push_back(std::move(evt));
+    }
+    return result;
 }
 
 statuses_t env_stack_t::get_last_statuses() const {
@@ -1340,7 +1357,9 @@ maybe_t<env_var_t> env_stack_t::get(const wcstring &key, env_mode_flags_t mode)
     return acquire_impl()->get(key, mode);
 }
 
-wcstring_list_t env_stack_t::get_names(int flags) const { return acquire_impl()->get_names(flags); }
+wcstring_list_t env_stack_t::get_names(env_mode_flags_t flags) const {
+    return acquire_impl()->get_names(flags);
+}
 
 int env_stack_t::set(const wcstring &key, env_mode_flags_t mode, wcstring_list_t vals) {
     // Historical behavior.
@@ -1365,9 +1384,9 @@ int env_stack_t::set(const wcstring &key, env_mode_flags_t mode, wcstring_list_t
             env_dispatch_var_change(key, *this);
         }
     }
-    // If the principal stack modified universal variables, then post a barrier.
-    if (ret.uvar_modified && is_principal()) {
-        universal_barrier();
+    // Mark if we modified a uvar.
+    if (ret.uvar_modified) {
+        s_uvars_locally_modified = true;
     }
     return ret.status;
 }
@@ -1390,8 +1409,8 @@ int env_stack_t::remove(const wcstring &key, int mode) {
             env_dispatch_var_change(key, *this);
         }
     }
-    if (ret.uvar_modified && is_principal()) {
-        universal_barrier();
+    if (ret.uvar_modified) {
+        s_uvars_locally_modified = true;
     }
     return ret.status;
 }
