@@ -5,14 +5,17 @@
 #include <stddef.h>
 #include <stdint.h>
 
+#include <map>
 #include <memory>
 #include <string>
 #include <unordered_map>
+#include <utility>
 #include <vector>
 
 #include "common.h"
 #include "maybe.h"
-#include "null_terminated_array.h"
+
+class owning_null_terminated_array_t;
 
 extern size_t read_byte_limit;
 extern bool curses_initialized;
@@ -20,7 +23,7 @@ extern bool curses_initialized;
 struct event_t;
 
 // Flags that may be passed as the 'mode' in env_stack_t::set() / environment_t::get().
-enum {
+enum : uint16_t {
     /// Default mode. Used with `env_stack_t::get()` to indicate the caller doesn't care what scope
     /// the var is in or whether it is exported or unexported.
     ENV_DEFAULT = 0,
@@ -45,7 +48,7 @@ enum {
     /// to modify a var that should not be modified by direct user action; e.g., a read-only var.
     ENV_USER = 1 << 8,
 };
-typedef uint32_t env_mode_flags_t;
+using env_mode_flags_t = uint16_t;
 
 /// Return values for `env_stack_t::set()`.
 enum { ENV_OK, ENV_PERM, ENV_SCOPE, ENV_INVALID, ENV_NOT_FOUND };
@@ -187,7 +190,7 @@ class environment_t {
    public:
     virtual maybe_t<env_var_t> get(const wcstring &key,
                                    env_mode_flags_t mode = ENV_DEFAULT) const = 0;
-    virtual wcstring_list_t get_names(int flags) const = 0;
+    virtual wcstring_list_t get_names(env_mode_flags_t flags) const = 0;
     virtual ~environment_t();
 
     /// Returns the PWD with a terminating slash.
@@ -201,7 +204,7 @@ class null_environment_t : public environment_t {
     ~null_environment_t() override;
 
     maybe_t<env_var_t> get(const wcstring &key, env_mode_flags_t mode = ENV_DEFAULT) const override;
-    wcstring_list_t get_names(int flags) const override;
+    wcstring_list_t get_names(env_mode_flags_t flags) const override;
 };
 
 /// A mutable environment which allows scopes to be pushed and popped.
@@ -229,7 +232,7 @@ class env_stack_t final : public environment_t {
     maybe_t<env_var_t> get(const wcstring &key, env_mode_flags_t mode = ENV_DEFAULT) const override;
 
     /// Implementation of environment_t.
-    wcstring_list_t get_names(int flags) const override;
+    wcstring_list_t get_names(env_mode_flags_t flags) const override;
 
     /// Sets the variable with the specified name to the given values.
     int set(const wcstring &key, env_mode_flags_t mode, wcstring_list_t vals);
@@ -259,10 +262,6 @@ class env_stack_t final : public environment_t {
     /// Pop the variable stack. Used for implementing local variables for functions and for-loops.
     void pop();
 
-    /// Synchronizes all universal variable changes: writes everything out, reads stuff in.
-    /// \return true if something changed, false otherwise.
-    bool universal_barrier();
-
     /// Returns an array containing all exported variables in a format suitable for execv.
     std::shared_ptr<owning_null_terminated_array_t> export_arr();
 
@@ -284,6 +283,12 @@ class env_stack_t final : public environment_t {
     /// Slightly optimized implementation.
     wcstring get_pwd_slash() const override;
 
+    /// Synchronizes universal variable changes.
+    /// If \p always is set, perform synchronization even if there's no pending changes from this
+    /// instance (that is, look for changes from other fish instances).
+    /// \return a list of events for changed variables.
+    std::vector<event_t> universal_sync(bool always);
+
     // Compatibility hack; access the "environment stack" from back when there was just one.
     static const std::shared_ptr<env_stack_t> &principal_ref();
     static env_stack_t &principal() { return *principal_ref(); }
@@ -297,10 +302,6 @@ bool get_use_posix_spawn();
 
 extern bool term_has_xn;  // does the terminal have the "eat_newline_glitch"
 
-/// Synchronizes all universal variable changes: writes everything out, reads stuff in.
-/// \return true if any value changed.
-bool env_universal_barrier();
-
 /// Returns true if we think the terminal supports setting its title.
 bool term_supports_setting_title();
 
@@ -313,4 +314,7 @@ wcstring env_get_runtime_path();
 void setenv_lock(const char *name, const char *value, int overwrite);
 void unsetenv_lock(const char *name);
 
+/// Returns the originally inherited variables and their values.
+/// This is a simple key->value map and not e.g. cut into paths.
+const std::map<wcstring, wcstring> &env_get_inherited();
 #endif
