@@ -4,20 +4,19 @@
 #include <errno.h>
 #include <limits.h>
 #include <locale.h>
-#include <stddef.h>
-#include <stdlib.h>
 #include <time.h>
 #include <unistd.h>
 
+#include <cstdlib>
 #include <cstring>
 #include <cwchar>
 
 #if HAVE_CURSES_H
-#include <curses.h>
+#include <curses.h>  // IWYU pragma: keep
 #elif HAVE_NCURSES_H
-#include <ncurses.h>
+#include <ncurses.h>  // IWYU pragma: keep
 #elif HAVE_NCURSES_CURSES_H
-#include <ncurses/curses.h>
+#include <ncurses/curses.h>  // IWYU pragma: keep
 #endif
 #if HAVE_TERM_H
 #include <term.h>
@@ -25,8 +24,6 @@
 #include <ncurses/term.h>
 #endif
 
-#include <assert.h>
-
 #include <algorithm>
 #include <functional>
 #include <memory>
@@ -38,8 +35,6 @@
 #include "complete.h"
 #include "env.h"
 #include "env_dispatch.h"
-#include "env_universal_common.h"
-#include "event.h"
 #include "fallback.h"  // IWYU pragma: keep
 #include "flog.h"
 #include "function.h"
@@ -48,29 +43,26 @@
 #include "input_common.h"
 #include "maybe.h"
 #include "output.h"
-#include "parser.h"
 #include "proc.h"
 #include "reader.h"
 #include "screen.h"
 #include "termsize.h"
 #include "trace.h"
-#include "wutil.h"  // IWYU pragma: keep
+#include "wcstringutil.h"
+#include "wutil.h"
 
-#define DEFAULT_TERM1 "ansi"
-#define DEFAULT_TERM2 "dumb"
+// Limit `read` to 100 MiB (bytes not wide chars) by default. This can be overridden by the
+// fish_read_limit variable.
+constexpr size_t DEFAULT_READ_BYTE_LIMIT = 100 * 1024 * 1024;
+size_t read_byte_limit = DEFAULT_READ_BYTE_LIMIT;
 
 /// List of all locale environment variable names that might trigger (re)initializing the locale
 /// subsystem. These are only the variables we're possibly interested in.
-static const wcstring locale_variables[] = {L"LANG",
-                                            L"LANGUAGE",
-                                            L"LC_ALL",
-                                            L"LC_COLLATE",
-                                            L"LC_CTYPE",
-                                            L"LC_MESSAGES",
-                                            L"LC_NUMERIC",
-                                            L"LC_TIME",
-                                            L"fish_allow_singlebyte_locale",
-                                            L"LOCPATH"};
+static const wcstring locale_variables[] = {
+    L"LANG",       L"LANGUAGE", L"LC_ALL",
+    L"LC_COLLATE", L"LC_CTYPE", L"LC_MESSAGES",
+    L"LC_NUMERIC", L"LC_TIME",  L"fish_allow_singlebyte_locale",
+    L"LOCPATH"};
 
 /// List of all curses environment variable names that might trigger (re)initializing the curses
 /// subsystem.
@@ -195,26 +187,12 @@ static void guess_emoji_width(const environment_t &vars) {
 
 /// React to modifying the given variable.
 void env_dispatch_var_change(const wcstring &key, env_stack_t &vars) {
-    ASSERT_IS_MAIN_THREAD();
     // Do nothing if not yet fully initialized.
     if (!s_var_dispatch_table) return;
 
     s_var_dispatch_table->dispatch(key, vars);
 }
 
-// Trigger events due to a universal variable changing.
-void env_universal_callbacks(env_stack_t *stack, const callback_data_list_t &callbacks) {
-    for (const callback_data_t &cb : callbacks) {
-        env_dispatch_var_change(cb.key, *stack);
-
-        // TODO: eliminate this principal_parser. Need to rationalize how multiple threads work
-        // here.
-        event_t evt =
-            cb.is_erase() ? event_t::variable_erase(cb.key) : event_t::variable_set(cb.key);
-        event_fire(parser_t::principal_parser(), evt);
-    }
-}
-
 static void handle_fish_term_change(const env_stack_t &vars) {
     update_fish_color_support(vars);
     reader_schedule_prompt_repaint();
@@ -236,6 +214,13 @@ static void handle_fish_history_change(const env_stack_t &vars) {
     reader_change_history(history_session_id(vars));
 }
 
+static void handle_fish_cursor_selection_mode_change(const env_stack_t &vars) {
+    auto mode = vars.get(L"fish_cursor_selection_mode");
+    reader_change_cursor_selection_mode(mode && mode->as_string() == L"inclusive"
+                                            ? cursor_selection_mode_t::inclusive
+                                            : cursor_selection_mode_t::exclusive);
+}
+
 void handle_autosuggestion_change(const env_stack_t &vars) {
     reader_set_autosuggestion_enabled(vars);
 }
@@ -270,26 +255,19 @@ static relaxed_atomic_bool_t g_use_posix_spawn{false};
 
 bool get_use_posix_spawn() { return g_use_posix_spawn; }
 
-extern "C" {
-const char *gnu_get_libc_version();
-}
-
 static bool allow_use_posix_spawn() {
     // OpenBSD's posix_spawn returns status 127, instead of erroring with ENOEXEC, when faced with a
     // shebangless script. Disable posix_spawn on OpenBSD.
 #if defined(__OpenBSD__)
     return false;
-#endif
-    bool result = true;
-    // uClibc defines __GLIBC__.
-#if defined(__GLIBC__) && !defined(__UCLIBC__)
+#elif defined(__GLIBC__) && !defined(__UCLIBC__)  // uClibc defines __GLIBC__
     // Disallow posix_spawn entirely on glibc < 2.24.
     // See #8021.
-    if (!__GLIBC_PREREQ(2, 24)) {
-        result = false;
-    }
+    return __GLIBC_PREREQ(2, 24) ? true : false;
+#else                                             // !defined(__OpenBSD__)
+    return true;
 #endif
-    return result;
+    return true;
 }
 
 static void handle_fish_use_posix_spawn_change(const environment_t &vars) {
@@ -314,6 +292,8 @@ static void handle_read_limit_change(const environment_t &vars) {
         } else {
             read_byte_limit = limit;
         }
+    } else {
+        read_byte_limit = DEFAULT_READ_BYTE_LIMIT;
     }
 }
 
@@ -349,6 +329,8 @@ static std::unique_ptr<const var_dispatch_table_t> create_dispatch_table() {
     var_dispatch_table->add(L"TZ", handle_tz_change);
     var_dispatch_table->add(L"fish_use_posix_spawn", handle_fish_use_posix_spawn_change);
     var_dispatch_table->add(L"fish_trace", handle_fish_trace);
+    var_dispatch_table->add(L"fish_cursor_selection_mode",
+                            handle_fish_cursor_selection_mode_change);
 
     // This std::move is required to avoid a build error on old versions of libc++ (#5801),
     // but it causes a different warning under newer versions of GCC (observed under GCC 9.3.0,
@@ -445,7 +427,7 @@ static void update_fish_color_support(const environment_t &vars) {
                 FLOGF(term_support, L"Truecolor support: Enabling for st");
                 support_term24bit = true;
             } else if (auto vte = vars.get(L"VTE_VERSION")) {
-                if (fish_wcstod(vte->as_string().c_str(), nullptr) > 3600) {
+                if (fish_wcstod(vte->as_string(), nullptr) > 3600) {
                     FLOGF(term_support, L"Truecolor support: Enabling for VTE version %ls",
                           vte->as_string().c_str());
                     support_term24bit = true;
@@ -459,28 +441,80 @@ static void update_fish_color_support(const environment_t &vars) {
 }
 
 // Try to initialize the terminfo/curses subsystem using our fallback terminal name. Do not set
-// `TERM` to our fallback. We're only doing this in the hope of getting a minimally functional
+// `TERM` to our fallback. We're only doing this in the hope of getting a functional
 // shell. If we launch an external command that uses TERM it should get the same value we were
 // given, if any.
-static bool initialize_curses_using_fallback(const char *term) {
-    // If $TERM is already set to the fallback name we're about to use there isn't any point in
-    // seeing if the fallback name can be used.
-    auto &vars = env_stack_t::globals();
+static void initialize_curses_using_fallbacks(const environment_t &vars) {
+    // xterm-256color is the most used terminal type by a massive margin,
+    // especially counting terminals that are mostly compatible.
+    const wchar_t *const fallbacks[] = {L"xterm-256color", L"xterm", L"ansi", L"dumb"};
+
+    wcstring termstr = L"";
     auto term_var = vars.get(L"TERM");
-    if (term_var.missing_or_empty()) return false;
-
-    auto term_env = wcs2string(term_var->as_string());
-    if (term_env == DEFAULT_TERM1 || term_env == DEFAULT_TERM2) return false;
-
-    if (is_interactive_session()) FLOGF(warning, _(L"Using fallback terminal type '%s'."), term);
-
-    int err_ret;
-    if (setupterm(const_cast<char *>(term), STDOUT_FILENO, &err_ret) == OK) return true;
-    if (is_interactive_session()) {
-        FLOGF(warning, _(L"Could not set up terminal using the fallback terminal type '%s'."),
-              term);
+    if (!term_var.missing_or_empty()) {
+        termstr = term_var->as_string();
     }
-    return false;
+
+    for (const wchar_t *fallback : fallbacks) {
+        // If $TERM is already set to the fallback name we're about to use there isn't any point in
+        // seeing if the fallback name can be used.
+        if (termstr == fallback) {
+            continue;
+        }
+
+        int err_ret = 0;
+        std::string term = wcs2string(fallback);
+        bool success = (setupterm(&term[0], STDOUT_FILENO, &err_ret) == OK);
+
+        if (is_interactive_session()) {
+            if (success) {
+                FLOGF(warning, _(L"Using fallback terminal type '%s'."), term.c_str());
+            } else {
+                FLOGF(warning,
+                      _(L"Could not set up terminal using the fallback terminal type '%s'."),
+                      term.c_str());
+            }
+        }
+        if (success) {
+            break;
+        }
+    }
+}
+
+// Apply any platform-specific hacks to cur_term/
+static void apply_term_hacks(const environment_t &vars) {
+    UNUSED(vars);
+    // Be careful, variables like "enter_italics_mode" are #defined to dereference through cur_term.
+    // See #8876.
+    if (!cur_term) {
+        return;
+    }
+#ifdef __APPLE__
+    // Hack in missing italics and dim capabilities omitted from MacOS xterm-256color terminfo
+    // Helps Terminal.app/iTerm
+    wcstring term_prog;
+    if (auto var = vars.get(L"TERM_PROGRAM")) {
+        term_prog = var->as_string();
+    }
+    if (term_prog == L"Apple_Terminal" || term_prog == L"iTerm.app") {
+        const auto term = vars.get(L"TERM");
+        if (term && term->as_string() == L"xterm-256color") {
+            static char sitm_esc[] = "\x1B[3m";
+            static char ritm_esc[] = "\x1B[23m";
+            static char dim_esc[] = "\x1B[2m";
+
+            if (!enter_italics_mode) {
+                enter_italics_mode = sitm_esc;
+            }
+            if (!exit_italics_mode) {
+                exit_italics_mode = ritm_esc;
+            }
+            if (!enter_dim_mode) {
+                enter_dim_mode = dim_esc;
+            }
+        }
+    }
+#endif
 }
 
 /// This is a pretty lame heuristic for detecting terminals that do not support setting the
@@ -534,10 +568,10 @@ static void init_curses(const environment_t &vars) {
         }
     }
 
-    int err_ret;
+    int err_ret{0};
     if (setupterm(nullptr, STDOUT_FILENO, &err_ret) == ERR) {
-        auto term = vars.get(L"TERM");
         if (is_interactive_session()) {
+            auto term = vars.get(L"TERM");
             FLOGF(warning, _(L"Could not set up terminal."));
             if (term.missing_or_empty()) {
                 FLOGF(warning, _(L"TERM environment variable not set."));
@@ -548,11 +582,11 @@ static void init_curses(const environment_t &vars) {
             }
         }
 
-        if (!initialize_curses_using_fallback(DEFAULT_TERM1)) {
-            initialize_curses_using_fallback(DEFAULT_TERM2);
-        }
+        initialize_curses_using_fallbacks(vars);
     }
 
+    apply_term_hacks(vars);
+
     can_set_term_title = does_term_support_setting_title(vars);
     term_has_xn =
         tigetflag(const_cast<char *>("xenl")) == 1;  // does terminal have the eat_newline_glitch
@@ -632,7 +666,3 @@ static void init_locale(const environment_t &vars) {
 
 /// Returns true if we think the terminal supports setting its title.
 bool term_supports_setting_title() { return can_set_term_title; }
-
-// Limit `read` to 100 MiB (bytes not wide chars) by default. This can be overridden by the
-// fish_read_limit variable.
-size_t read_byte_limit = 100 * 1024 * 1024;
