@@ -4,14 +4,10 @@
 
 #include "config.h"  // IWYU pragma: keep
 
-#include <memory>
-
 #include "common.h"
-#include "env_universal_common.h"
 
 class environment_t;
 class env_stack_t;
-class parser_t;
 
 /// Initialize variable dispatch.
 void env_dispatch_init(const environment_t &vars);
