@@ -3,11 +3,8 @@
 
 #include "config.h"
 
-#include <stddef.h>
-#include <stdint.h>
-#include <unistd.h>
 // The following include must be kept despite what IWYU says. That's because of the interaction
-// between the weak linking of `wcsdup` and `wcscasecmp` via `#define`s below and the declarations
+// between the weak linking of `wcscasecmp` via `#define`s below and the declarations
 // in <wchar.h>. At least on OS X if we don't do this we get compilation errors do to the macro
 // substitution if wchar.h is included after this header.
 #include <cwchar>  // IWYU pragma: keep
@@ -77,14 +74,6 @@ char *tparm_solaris_kludge(char *str, long p1 = 0, long p2 = 0, long p3 = 0, lon
 
 /// These functions are missing from Solaris 10, and only accessible from
 /// Solaris 11 in the std:: namespace.
-#ifndef HAVE_WCSDUP
-#ifdef HAVE_STD__WCSDUP
-using std::wcsdup;
-#else
-wchar_t *wcsdup(const wchar_t *in);
-#endif  // HAVE_STD__WCSDUP
-#endif  // HAVE_WCSDUP
-
 #ifndef HAVE_WCSCASECMP
 #ifdef HAVE_STD__WCSCASECMP
 using std::wcscasecmp;
@@ -109,20 +98,6 @@ int wcsncasecmp(const wchar_t *s1, const wchar_t *s2, size_t n);
 #endif
 #endif
 
-#ifndef HAVE_WCSNDUP
-/// Fallback for wcsndup function. Returns a copy of \c in, truncated to a maximum length of \c c.
-wchar_t *wcsndup(const wchar_t *in, size_t c);
-#endif
-
-#ifndef HAVE_WCSLCPY
-/// Copy src to string dst of size siz.  At most siz-1 characters will be copied.  Always NUL
-/// terminates (unless siz == 0).  Returns std::wcslen(src); if retval >= siz, truncation occurred.
-///
-/// This is the OpenBSD strlcpy function, modified for wide characters, and renamed to reflect this
-/// change.
-size_t wcslcpy(wchar_t *dst, const wchar_t *src, size_t siz);
-#endif
-
 // autoconf may fail to detect gettext (645), so don't define a function call gettext or we'll get
 // build errors.
 
