@@ -4,6 +4,9 @@
 #include <chrono>
 #include <cstdint>
 #include <functional>
+#include <utility>
+#include <vector>
+
 // Needed for musl
 #include <sys/select.h>  // IWYU pragma: keep
 
@@ -11,8 +14,6 @@
 #include "fds.h"
 #include "maybe.h"
 
-class fd_monitor_t;
-
 /// Each item added to fd_monitor_t is assigned a unique ID, which is not recycled.
 /// Items may have their callback triggered immediately by passing the ID.
 /// Zero is a sentinel.
@@ -110,7 +111,7 @@ class fd_monitor_t {
     // Poke items in the pokelist, removing any items that close their FD.
     // The pokelist is consumed after this.
     // This is only called in the background thread.
-    void poke_in_background(poke_list_t pokelist);
+    void poke_in_background(const poke_list_t &pokelist);
 
     // The list of items to monitor. This is only accessed on the background thread.
     item_list_t items_{};
