@@ -23,18 +23,19 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 #include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <wctype.h>
 
+#include <algorithm>
+#include <cstdint>
 #include <cstring>
 #include <cwchar>
+#include <cwctype>
 #include <memory>
-#include <stack>
 #include <string>
-#include <tuple>
+#include <type_traits>
+#include <utility>
 #include <vector>
 
 #include "ast.h"
-#include "color.h"
 #include "common.h"
 #include "env.h"
 #include "expand.h"
@@ -42,12 +43,14 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 #include "fish_version.h"
 #include "flog.h"
 #include "future_feature_flags.h"
+#include "global_safety.h"
 #include "highlight.h"
+#include "maybe.h"
 #include "operation_context.h"
-#include "output.h"
 #include "parse_constants.h"
 #include "parse_util.h"
 #include "print_help.h"
+#include "tokenizer.h"
 #include "wcstringutil.h"
 #include "wutil.h"  // IWYU pragma: keep
 
@@ -912,7 +915,7 @@ int main(int argc, char *argv[]) {
                                        {"ansi", no_argument, nullptr, 2},
                                        {"pygments", no_argument, nullptr, 3},
                                        {"check", no_argument, nullptr, 'c'},
-                                       {nullptr, 0, nullptr, 0}};
+                                       {}};
 
     int opt;
     while ((opt = getopt_long(argc, argv, short_opts, long_opts, nullptr)) != -1) {
