@@ -8,10 +8,7 @@
 // Type "exit" or "quit" to terminate the program.
 #include "config.h"  // IWYU pragma: keep
 
-#include <errno.h>
 #include <getopt.h>
-#include <signal.h>
-#include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <termios.h>
@@ -19,7 +16,6 @@
 
 #include <cstring>
 #include <cwchar>
-#include <memory>
 #include <string>
 #include <vector>
 
@@ -29,6 +25,7 @@
 #include "fish_version.h"
 #include "input.h"
 #include "input_common.h"
+#include "maybe.h"
 #include "parser.h"
 #include "print_help.h"
 #include "proc.h"
@@ -42,7 +39,7 @@ static const wchar_t *ctrl_symbolic_names[] = {
     nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, L"\\a",
     L"\\b",  L"\\t",  L"\\n",  L"\\v",  L"\\f",  L"\\r",  nullptr, nullptr,
     nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr,
-    nullptr, nullptr, nullptr, L"\\e",  nullptr, nullptr, nullptr, nullptr};
+    nullptr, nullptr, nullptr, L"\\e",  L"\\x1c", nullptr, nullptr, nullptr};
 
 /// Return true if the recent sequence of characters indicates the user wants to exit the program.
 static bool should_exit(wchar_t wc) {
@@ -230,7 +227,7 @@ static void process_input(bool continuous_mode, bool verbose) {
     std::vector<wchar_t> bind_chars;
 
     std::fwprintf(stderr, L"Press a key:\n");
-    for (;;) {
+    while (!check_exit_loop_maybe_warning(nullptr)) {
         maybe_t<char_event_t> evt{};
         if (reader_test_and_clear_interrupted()) {
             evt = char_event_t{shell_modes.c_cc[VINTR]};
@@ -303,7 +300,7 @@ static bool parse_flags(int argc, char **argv, bool *continuous_mode, bool *verb
                                        {"help", no_argument, nullptr, 'h'},
                                        {"version", no_argument, nullptr, 'v'},
                                        {"verbose", no_argument, nullptr, 'V'},
-                                       {nullptr, 0, nullptr, 0}};
+                                       {}};
     int opt;
     bool error = false;
     while (!error && (opt = getopt_long(argc, argv, short_opts, long_opts, nullptr)) != -1) {
