@@ -4,20 +4,20 @@
 
 #include "config.h"  // IWYU pragma: keep
 
+#include <errno.h>
 #include <stdio.h>
 
+#include <cstdint>
 #include <string>
 #include <type_traits>
-#include <utility>
+#include <vector>
 
+#include "common.h"
 #include "global_safety.h"
 
 using wcstring = std::wstring;
 using wcstring_list_t = std::vector<wcstring>;
 
-template <typename T>
-class owning_lock;
-
 namespace flog_details {
 
 class category_list_t;
@@ -109,6 +109,8 @@ class category_list_t {
     category_t path{L"path", L"Searching/using paths"};
 
     category_t screen{L"screen", L"Screen repaints"};
+
+    category_t abbrs{L"abbrs", L"Abbreviation expansion"};
 };
 
 /// The class responsible for logging.
