@@ -1,38 +1,31 @@
 // History functions, part of the user interface.
 #include "config.h"  // IWYU pragma: keep
 
-#include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <pthread.h>
-#include <stddef.h>
 #include <stdio.h>
-#include <stdlib.h>
 
-#include <cstdint>
 #include <cstring>
 // We need the sys/file.h for the flock() declaration on Linux but not OS X.
 #include <sys/file.h>  // IWYU pragma: keep
 #include <sys/stat.h>
 #include <unistd.h>
-#include <wctype.h>
 
 #include <algorithm>
-#include <atomic>
 #include <chrono>
 #include <cwchar>
 #include <functional>
 #include <iterator>
 #include <map>
-#include <numeric>
 #include <random>
-#include <type_traits>
 #include <unordered_set>
 
 #include "ast.h"
 #include "common.h"
 #include "env.h"
+#include "expand.h"
 #include "fallback.h"  // IWYU pragma: keep
+#include "fds.h"
 #include "flog.h"
 #include "global_safety.h"
 #include "history.h"
@@ -40,11 +33,10 @@
 #include "io.h"
 #include "iothread.h"
 #include "lru.h"
+#include "operation_context.h"
 #include "parse_constants.h"
 #include "parse_util.h"
-#include "parser.h"
 #include "path.h"
-#include "reader.h"
 #include "wcstringutil.h"
 #include "wildcard.h"  // IWYU pragma: keep
 #include "wutil.h"     // IWYU pragma: keep
@@ -122,34 +114,11 @@ maybe_t<wcstring> history_filename(const wcstring &session_id, const wcstring &s
     return result;
 }
 
-/// Lock the history file.
-/// Returns true on success, false on failure.
-bool history_file_lock(int fd, int lock_type) {
-    static std::atomic<bool> do_locking(true);
-    if (!do_locking) return false;
-
-    double start_time = timef();
-    int retval = flock(fd, lock_type);
-    double duration = timef() - start_time;
-    if (duration > 0.25) {
-        FLOGF(warning, _(L"Locking the history file took too long (%.3f seconds)."), duration);
-        // we've decided to stop doing any locking behavior
-        // but make sure we don't leave the file locked!
-        if (retval == 0 && lock_type != LOCK_UN) {
-            flock(fd, LOCK_UN);
-        }
-        do_locking = false;
-        return false;
-    }
-    return retval != -1;
-}
-
 }  // anonymous namespace
 
-class history_lru_cache_t : public lru_cache_t<history_lru_cache_t, history_item_t> {
+class history_lru_cache_t : public lru_cache_t<history_item_t> {
    public:
-    explicit history_lru_cache_t(size_t max)
-        : lru_cache_t<history_lru_cache_t, history_item_t>(max) {}
+    explicit history_lru_cache_t(size_t max) : lru_cache_t<history_item_t>(max) {}
 
     /// Function to add a history item.
     void add_item(history_item_t item) {
@@ -238,7 +207,7 @@ struct history_impl_t {
     // item_at_index until a call to resolve_pending(). Pending items are tracked with an offset
     // into the array of new items, so adding a non-pending item has the effect of resolving all
     // pending items.
-    void add(history_item_t item, bool pending = false, bool do_save = true);
+    void add(history_item_t &&item, bool pending = false, bool do_save = true);
 
     // Internal function.
     void clear_file_state();
@@ -262,7 +231,9 @@ struct history_impl_t {
     uint32_t disable_automatic_save_counter{0};
 
     // Deleted item contents.
-    std::unordered_set<wcstring> deleted_items{};
+    // Boolean describes if it should be deleted only in this session or in all
+    // (used in deduplication).
+    std::unordered_map<wcstring, bool> deleted_items{};
 
     // The buffer containing the history file contents.
     std::unique_ptr<history_file_contents_t> file_contents{};
@@ -288,6 +259,10 @@ struct history_impl_t {
     // List of old items, as offsets into out mmap data.
     std::deque<size_t> old_item_offsets{};
 
+    // If set, we gave up on file locking because it took too long.
+    // Note this is shared among all history instances.
+    static relaxed_atomic_bool_t abandoned_locking;
+
     /// \return a timestamp for new items - see the implementation for a subtlety.
     time_t timestamp_now() const;
 
@@ -350,6 +325,9 @@ struct history_impl_t {
     // Irreversibly clears history.
     void clear();
 
+    // Clears only session.
+    void clear_session();
+
     // Populates from older location ()in config path, rather than data path).
     void populate_from_config_path();
 
@@ -377,9 +355,39 @@ struct history_impl_t {
 
     // Return the number of history entries.
     size_t size();
+
+    // Maybe lock a history file.
+    // \return true if successful, false if locking was skipped.
+    static bool maybe_lock_file(int fd, int lock_type);
+    static void unlock_file(int fd);
 };
 
-void history_impl_t::add(history_item_t item, bool pending, bool do_save) {
+relaxed_atomic_bool_t history_impl_t::abandoned_locking{false};
+
+// static
+bool history_impl_t::maybe_lock_file(int fd, int lock_type) {
+    assert(!(lock_type & LOCK_UN) && "Do not use lock_file to unlock");
+
+    // Don't lock if it took too long before, if we are simulating a failing lock, or if our history
+    // is on a remote filesystem.
+    if (abandoned_locking) return false;
+    if (history_t::chaos_mode) return false;
+    if (path_get_data_remoteness() == dir_remoteness_t::remote) return false;
+
+    double start_time = timef();
+    int retval = flock(fd, lock_type);
+    double duration = timef() - start_time;
+    if (duration > 0.25) {
+        FLOGF(warning, _(L"Locking the history file took too long (%.3f seconds)."), duration);
+        abandoned_locking = true;
+    }
+    return retval != -1;
+}
+
+// static
+void history_impl_t::unlock_file(int fd) { flock(fd, LOCK_UN); }
+
+void history_impl_t::add(history_item_t &&item, bool pending, bool do_save) {
     assert(item.timestamp() != 0 && "Should not add an item with a 0 timestamp");
     // We use empty items as sentinels to indicate the end of history.
     // Do not allow them to be added (#6032).
@@ -441,7 +449,7 @@ void history_impl_t::save_unless_disabled() {
 // case-sensitive, matches.
 void history_impl_t::remove(const wcstring &str_to_remove) {
     // Add to our list of deleted items.
-    deleted_items.insert(str_to_remove);
+    deleted_items.insert(std::pair<wcstring, bool>(str_to_remove, false));
 
     size_t idx = new_items.size();
     while (idx--) {
@@ -574,7 +582,9 @@ void history_impl_t::populate_from_file_contents() {
     old_item_offsets.clear();
     if (file_contents) {
         size_t cursor = 0;
-        while (auto offset = file_contents->offset_of_next_item(&cursor, boundary_timestamp)) {
+        maybe_t<size_t> offset;
+        while ((offset =
+                    file_contents->offset_of_next_item(&cursor, boundary_timestamp)).has_value()) {
             // Remember this item.
             old_item_offsets.push_back(*offset);
         }
@@ -592,20 +602,18 @@ void history_impl_t::load_old_if_needed() {
         autoclose_fd_t file{wopen_cloexec(*filename, O_RDONLY)};
         int fd = file.fd();
         if (fd >= 0) {
-            // Take a read lock to guard against someone else appending. This is released when the
-            // file is closed (below). We will read the file after releasing the lock, but that's
+            // Take a read lock to guard against someone else appending. This is released after
+            // getting the file's length. We will read the file after releasing the lock, but that's
             // not a problem, because we never modify already written data. In short, the purpose of
             // this lock is to ensure we don't see the file size change mid-update.
             //
             // We may fail to lock (e.g. on lockless NFS - see issue #685. In that case, we proceed
             // as if it did not fail. The risk is that we may get an incomplete history item; this
             // is unlikely because we only treat an item as valid if it has a terminating newline.
-            //
-            // Simulate a failing lock in chaos_mode.
-            if (!history_t::chaos_mode) history_file_lock(fd, LOCK_SH);
+            bool locked = maybe_lock_file(fd, LOCK_SH);
             file_contents = history_file_contents_t::create(fd);
             this->history_file_id = file_contents ? file_id_for_fd(fd) : kInvalidFileID;
-            if (!history_t::chaos_mode) history_file_lock(fd, LOCK_UN);
+            if (locked) unlock_file(fd);
 
             time_profiler_t profiler("populate_from_file_contents");  //!OCLINT(side-effect)
             this->populate_from_file_contents();
@@ -613,14 +621,24 @@ void history_impl_t::load_old_if_needed() {
     }
 }
 
-bool history_search_t::go_backwards() {
+bool history_search_t::go_to_next_match(history_search_direction_t direction) {
     // Backwards means increasing our index.
-    const auto max_index = static_cast<size_t>(-1);
+    size_t invalid_index;
+    ssize_t increment;
 
-    if (current_index_ == max_index) return false;
+    if (direction == history_search_direction_t::backward) {
+        invalid_index = static_cast<size_t>(-1);
+        increment = 1;
+    } else {
+        assert(direction == history_search_direction_t::forward);
+        invalid_index = 0;
+        increment = -1;
+    }
+
+    if (current_index_ == invalid_index) return false;
 
     size_t index = current_index_;
-    while (++index < max_index) {
+    while ((index += increment) != invalid_index) {
         history_item_t item = history_->item_at_index(index);
 
         // We're done if it's empty or we cancelled.
@@ -653,6 +671,8 @@ const history_item_t &history_search_t::current_item() const {
 
 const wcstring &history_search_t::current_string() const { return this->current_item().str(); }
 
+size_t history_search_t::current_index() const { return this->current_index_; }
+
 void history_impl_t::clear_file_state() {
     // Erase everything we know about our file.
     file_contents.reset();
@@ -707,15 +727,27 @@ bool history_impl_t::rewrite_to_temporary_file(int existing_fd, int dst_fd) cons
     // old file contents).
     if (auto local_file = history_file_contents_t::create(existing_fd)) {
         size_t cursor = 0;
-        while (auto offset = local_file->offset_of_next_item(&cursor, 0)) {
+        maybe_t<size_t> offset;
+        while ((offset = local_file->offset_of_next_item(&cursor, 0)).has_value()) {
             // Try decoding an old item.
             history_item_t old_item = local_file->decode_item(*offset);
 
-            if (old_item.empty() || deleted_items.count(old_item.str()) > 0) {
-                continue;
+            // If old item is newer than session always erase if in deleted.
+            if (old_item.timestamp() > boundary_timestamp) {
+                if (old_item.empty() || deleted_items.count(old_item.str()) > 0) {
+                    continue;
+                }
+                lru.add_item(std::move(old_item));
+            } else {
+                // If old item is older and in deleted items don't erase if added by
+                // clear_session.
+                if (old_item.empty() || (deleted_items.count(old_item.str()) > 0 &&
+                                         !deleted_items.at(old_item.str()))) {
+                    continue;
+                }
+                // Add this old item.
+                lru.add_item(std::move(old_item));
             }
-            // Add this old item.
-            lru.add_item(std::move(old_item));
         }
     }
 
@@ -816,10 +848,11 @@ bool history_impl_t::save_internal_via_rewrite() {
         autoclose_fd_t target_fd_after{wopen_cloexec(target_name, O_RDONLY)};
         if (target_fd_after.valid()) {
             // critical to take the lock before checking file IDs,
-            // and hold it until after we are done replacing
-            // Also critical to check the file at the path, NOT based on our fd
-            // It's only OK to replace the file while holding the lock
-            history_file_lock(target_fd_after.fd(), LOCK_EX);
+            // and hold it until after we are done replacing.
+            // Also critical to check the file at the path, NOT based on our fd.
+            // It's only OK to replace the file while holding the lock.
+            // Note any lock is released when target_fd_after is closed.
+            (void)maybe_lock_file(target_fd_after.fd(), LOCK_EX);
             new_file_id = file_id_for_path(target_name);
         }
         bool can_replace_file = (new_file_id == orig_file_id || new_file_id == kInvalidFileID);
@@ -913,9 +946,7 @@ bool history_impl_t::save_internal_via_appending() {
         // we may get interleaved history items, which is considered better than no history, or
         // forcing everything through the slow copy-move mode. We try to minimize this possibility
         // by writing with O_APPEND.
-        //
-        // Simulate a failing lock in chaos_mode
-        if (!history_t::chaos_mode) history_file_lock(fd.fd(), LOCK_EX);
+        maybe_lock_file(fd.fd(), LOCK_EX);
         const file_id_t file_id = file_id_for_fd(fd.fd());
         if (file_id_for_path(history_path) == file_id) {
             // File IDs match, so the file we opened is still at that path
@@ -1064,6 +1095,15 @@ void history_impl_t::clear() {
     this->clear_file_state();
 }
 
+void history_impl_t::clear_session() {
+    for (const auto &item : new_items) {
+        deleted_items.insert(std::pair<wcstring, bool>(item.str(), true));
+    }
+
+    new_items.clear();
+    first_unwritten_new_item_index = 0;
+}
+
 bool history_impl_t::is_default() const { return name == DFLT_FISH_HISTORY_SESSION_ID; }
 
 bool history_impl_t::is_empty() {
@@ -1096,7 +1136,7 @@ bool history_impl_t::is_empty() {
 
 void history_t::add(wcstring str) {
     auto imp = this->impl();
-    time_t when =  imp->timestamp_now();
+    time_t when = imp->timestamp_now();
     imp->add(history_item_t(std::move(str), when));
 }
 
@@ -1139,34 +1179,32 @@ void history_impl_t::populate_from_config_path() {
 static bool should_import_bash_history_line(const wcstring &line) {
     if (line.empty()) return false;
 
-    if (ast::ast_t::parse(line).errored()) return false;
-
-    // In doing this test do not allow incomplete strings. Hence the "false" argument.
-    parse_error_list_t errors;
-    parse_util_detect_errors(line, &errors);
-    if (!errors.empty()) return false;
-
     // The following are Very naive tests!
 
     // Skip comments.
     if (line[0] == '#') return false;
 
-    // Skip lines with backticks.
-    if (line.find('`') != std::string::npos) return false;
+    // Skip lines with backticks because we don't have that syntax,
+    // Skip brace expansions and globs because they don't work like ours
+    // Skip lines with literal tabs since we don't handle them well and we don't know what they
+    // mean. It could just be whitespace or it's actually passed somewhere (like e.g. `sed`).
+    // Skip lines that end with a backslash. We do not handle multiline commands from bash history.
+    if (line.find_first_of(L"`{*\t\\") != std::string::npos) return false;
 
     // Skip lines with [[...]] and ((...)) since we don't handle those constructs.
     if (line.find(L"[[") != std::string::npos) return false;
     if (line.find(L"]]") != std::string::npos) return false;
     if (line.find(L"((") != std::string::npos) return false;
     if (line.find(L"))") != std::string::npos) return false;
-    // Skip lines with literal tabs since we don't handle them well and we don't know what they
-    // mean. It could just be whitespace or it's actually passed somewhere (like e.g. `sed`).
-    if (line.find(L'\t') != std::string::npos) return false;
+    // "<<" here is a proxy for heredocs (and herestrings).
+    if (line.find(L"<<") != std::string::npos) return false;
 
-    // Skip lines that end with a backslash. We do not handle multiline commands from bash history.
-    if (line.back() == L'\\') return false;
+    if (ast::ast_t::parse(line).errored()) return false;
 
-    return true;
+    // In doing this test do not allow incomplete strings. Hence the "false" argument.
+    parse_error_list_t errors;
+    parse_util_detect_errors(line, &errors);
+    return errors.empty();
 }
 
 /// Import a bash command history file. Bash's history format is very simple: just lines with #s for
@@ -1240,8 +1278,6 @@ wcstring history_session_id(const environment_t &vars) {
         wcstring session_id = var->as_string();
         if (session_id.empty()) {
             result.clear();
-        } else if (session_id == L"default") {
-            // using the default value
         } else if (valid_var_name(session_id)) {
             result = session_id;
         } else {
@@ -1313,7 +1349,15 @@ struct history_t::impl_wrapper_t {
 void history_impl_t::resolve_pending() { this->has_pending_item = false; }
 
 bool history_t::chaos_mode = false;
+
+/* OpenBSD's mmap is not synchronized with other file operations. In particular it appears we may
+ * write() a file, fsync() it, close it, mmap() it, and call msync(), and we still may not see the
+ * newly written data. Just don't try mmap here. */
+#if defined(__OpenBSD__)
+bool history_t::never_mmap = true;
+#else
 bool history_t::never_mmap = false;
+#endif
 
 history_t::history_t(wcstring name) : wrap_(make_unique<impl_wrapper_t>(std::move(name))) {}
 
@@ -1327,7 +1371,7 @@ bool history_t::is_default() const { return impl()->is_default(); }
 
 bool history_t::is_empty() { return impl()->is_empty(); }
 
-void history_t::add(history_item_t item, bool pending) { impl()->add(std::move(item), pending); }
+void history_t::add(history_item_t &&item, bool pending) { impl()->add(std::move(item), pending); }
 
 void history_t::remove(const wcstring &str) { impl()->remove(str); }
 
@@ -1423,7 +1467,7 @@ static void do_1_history_search(history_t *hist, history_search_type_t search_ty
                                 const cancel_checker_t &cancel_check) {
     history_search_t searcher = history_search_t(hist, search_string, search_type,
                                                  case_sensitive ? 0 : history_search_ignore_case);
-    while (!cancel_check() && searcher.go_backwards()) {
+    while (!cancel_check() && searcher.go_to_next_match(history_search_direction_t::backward)) {
         if (!func(searcher.current_item())) {
             break;
         }
@@ -1438,8 +1482,10 @@ bool history_t::search(history_search_type_t search_type, const wcstring_list_t
     wcstring_list_t collected;
     wcstring formatted_record;
     size_t remaining = max_items;
+    bool output_error = false;
 
-    // The function we use to act on each item.
+    // The function we use to act on each item. The return value indicates whether the search should
+    // continue (true) or stop (on false).
     std::function<bool(const history_item_t &item)> func = [&](const history_item_t &item) -> bool {
         if (remaining == 0) return false;
         remaining -= 1;
@@ -1449,7 +1495,11 @@ bool history_t::search(history_search_type_t search_type, const wcstring_list_t
             collected.push_back(std::move(formatted_record));
         } else {
             // We can output this immediately.
-            streams.out.append(formatted_record);
+            if (!streams.out.append(formatted_record)) {
+                // This can happen if the user hit Ctrl-C to abort (maybe after the first page?).
+                output_error = true;
+                return false;
+            }
         }
         return true;
     };
@@ -1470,14 +1520,23 @@ bool history_t::search(history_search_type_t search_type, const wcstring_list_t
     }
 
     // Output any items we collected (which only happens in reverse).
-    for (auto iter = collected.rbegin(); iter != collected.rend(); ++iter) {
-        streams.out.append(*iter);
+    for (auto iter = collected.rbegin(); !output_error && iter != collected.rend(); ++iter) {
+        if (!streams.out.append(*iter)) {
+            // Don't force an error if output was aborted (typically via Ctrl-C/SIGINT); just don't
+            // try writing any more.
+            output_error = true;
+        }
     }
+
+    // We are intentionally not returning false in case of an output error, as the user aborting the
+    // output early (the most common case) isn't a reason to exit w/ a non-zero status code.
     return true;
 }
 
 void history_t::clear() { impl()->clear(); }
 
+void history_t::clear_session() { impl()->clear_session(); }
+
 void history_t::populate_from_config_path() { impl()->populate_from_config_path(); }
 
 void history_t::populate_from_bash(FILE *f) { impl()->populate_from_bash(f); }
