@@ -7,12 +7,13 @@
 #include "parse_util.h"
 
 #include <stdarg.h>
-#include <stdlib.h>
 
+#include <algorithm>
 #include <cwchar>
+#include <cwctype>
 #include <memory>
 #include <string>
-#include <type_traits>
+#include <utility>
 
 #include "ast.h"
 #include "builtin.h"
@@ -20,9 +21,9 @@
 #include "expand.h"
 #include "fallback.h"  // IWYU pragma: keep
 #include "future_feature_flags.h"
+#include "operation_context.h"
 #include "parse_constants.h"
-#include "parse_util.h"
-#include "parser.h"
+#include "parse_tree.h"
 #include "tokenizer.h"
 #include "wcstringutil.h"
 #include "wildcard.h"
@@ -39,6 +40,9 @@
 /// Error message for arguments to 'end'
 #define END_ARG_ERR_MSG _(L"'end' does not take arguments. Did you forget a ';'?")
 
+/// Error message when 'time' is in a pipeline.
+#define TIME_IN_PIPELINE_ERR_MSG _(L"The 'time' command may only be at the beginning of a pipeline")
+
 /// Maximum length of a variable name to show in error reports before truncation
 static constexpr int var_err_len = 16;
 
@@ -88,9 +92,9 @@ size_t parse_util_get_offset(const wcstring &str, int line, long line_offset) {
 }
 
 static int parse_util_locate_cmdsub(const wchar_t *in, const wchar_t **begin, const wchar_t **end,
-                                    bool allow_incomplete, bool *inout_is_quoted) {
+                                    bool allow_incomplete, bool *inout_is_quoted,
+                                    bool *out_has_dollar) {
     bool escaped = false;
-    bool is_first = true;
     bool is_token_begin = true;
     bool syntax_error = false;
     int paran_count = 0;
@@ -101,13 +105,15 @@ static int parse_util_locate_cmdsub(const wchar_t *in, const wchar_t **begin, co
     assert(in && "null parameter");
 
     const wchar_t *pos = in;
+    const wchar_t *last_dollar = nullptr;
     auto process_opening_quote = [&](wchar_t quote) -> bool /* ok */ {
         const wchar_t *q_end = quote_end(pos, quote);
         if (!q_end) return false;
         if (*q_end == L'$') {
+            last_dollar = q_end;
             quoted_cmdsubs.push_back(paran_count);
         }
-        // We want to report whether the outermost comand substitution between
+        // We want to report whether the outermost command substitution between
         // paran_begin..paran_end is quoted.
         if (paran_count == 0 && inout_is_quoted) {
             *inout_is_quoted = *q_end == L'$';
@@ -128,10 +134,15 @@ static int parse_util_locate_cmdsub(const wchar_t *in, const wchar_t **begin, co
                 escaped = true;
             } else if (*pos == L'#' && is_token_begin) {
                 pos = comment_end(pos) - 1;
+            } else if (*pos == L'$') {
+                last_dollar = pos;
             } else {
                 if (*pos == L'(') {
                     if ((paran_count == 0) && (paran_begin == nullptr)) {
                         paran_begin = pos;
+                        if (out_has_dollar) {
+                            *out_has_dollar = last_dollar == pos - 1;
+                        }
                     }
 
                     paran_count++;
@@ -148,7 +159,7 @@ static int parse_util_locate_cmdsub(const wchar_t *in, const wchar_t **begin, co
                         break;
                     }
 
-                    // Check if the ) did complete a quoted command substituion.
+                    // Check if the ) did complete a quoted command substitution.
                     if (!quoted_cmdsubs.empty() && quoted_cmdsubs.back() == paran_count) {
                         quoted_cmdsubs.pop_back();
                         // Quoted command substitutions temporarily close double quotes.
@@ -167,12 +178,11 @@ static int parse_util_locate_cmdsub(const wchar_t *in, const wchar_t **begin, co
                     }
                 }
             }
-            is_token_begin = is_token_delimiter(pos[0], is_first, pos[1]);
+            is_token_begin = is_token_delimiter(pos[0], pos[1]);
         } else {
             escaped = false;
             is_token_begin = false;
         }
-        is_first = false;
     }
 
     syntax_error |= (paran_count < 0);
@@ -241,7 +251,8 @@ long parse_util_slice_length(const wchar_t *in) {
 
 int parse_util_locate_cmdsubst_range(const wcstring &str, size_t *inout_cursor_offset,
                                      wcstring *out_contents, size_t *out_start, size_t *out_end,
-                                     bool accept_incomplete, bool *inout_is_quoted) {
+                                     bool accept_incomplete, bool *inout_is_quoted,
+                                     bool *out_has_dollar) {
     // Clear the return values.
     if (out_contents != nullptr) out_contents->clear();
     *out_start = 0;
@@ -258,7 +269,7 @@ int parse_util_locate_cmdsubst_range(const wcstring &str, size_t *inout_cursor_o
     const wchar_t *bracket_range_end = nullptr;
 
     int ret = parse_util_locate_cmdsub(valid_range_start, &bracket_range_begin, &bracket_range_end,
-                                       accept_incomplete, inout_is_quoted);
+                                       accept_incomplete, inout_is_quoted, out_has_dollar);
     if (ret <= 0) {
         return ret;
     }
@@ -300,7 +311,7 @@ void parse_util_cmdsubst_extent(const wchar_t *buff, size_t cursor_pos, const wc
     const wchar_t *pos = buff;
     for (;;) {
         const wchar_t *begin = nullptr, *end = nullptr;
-        if (parse_util_locate_cmdsub(pos, &begin, &end, true, nullptr) <= 0) {
+        if (parse_util_locate_cmdsub(pos, &begin, &end, true, nullptr, nullptr) <= 0) {
             // No subshell found, all done.
             break;
         }
@@ -356,7 +367,7 @@ static void job_or_process_extent(bool process, const wchar_t *buff, size_t curs
     if (b) *b = end;
 
     const wcstring buffcpy(begin, end);
-    tokenizer_t tok(buffcpy.c_str(), TOK_ACCEPT_UNFINISHED);
+    tokenizer_t tok(buffcpy.c_str(), TOK_ACCEPT_UNFINISHED | TOK_SHOW_COMMENTS);
     maybe_t<tok_t> token{};
     while ((token = tok.next()) && !finished) {
         size_t tok_begin = token->offset;
@@ -366,13 +377,12 @@ static void job_or_process_extent(bool process, const wchar_t *buff, size_t curs
                 if (!process) {
                     break;
                 }
+                __fallthrough__
             }
-            /* FALLTHROUGH */
             case token_type_t::end:
             case token_type_t::background:
             case token_type_t::andand:
-            case token_type_t::oror:
-            case token_type_t::comment: {
+            case token_type_t::oror: {
                 if (tok_begin >= pos) {
                     finished = 1;
                     if (b) *b = const_cast<wchar_t *>(begin) + tok_begin;
@@ -544,7 +554,7 @@ wchar_t parse_util_get_quote_type(const wcstring &cmd, size_t pos) {
 wcstring parse_util_escape_string_with_quote(const wcstring &cmd, wchar_t quote, bool no_tilde) {
     wcstring result;
     if (quote == L'\0') {
-        escape_flags_t flags = ESCAPE_ALL | ESCAPE_NO_QUOTED | (no_tilde ? ESCAPE_NO_TILDE : 0);
+        escape_flags_t flags = ESCAPE_NO_QUOTED | (no_tilde ? ESCAPE_NO_TILDE : 0);
         result = escape_string(cmd, flags);
     } else {
         // Here we are going to escape a string with quotes.
@@ -800,11 +810,11 @@ std::vector<int> parse_util_compute_indents(const wcstring &src) {
 
 /// Append a syntax error to the given error list.
 static bool append_syntax_error(parse_error_list_t *errors, size_t source_location,
-                                const wchar_t *fmt, ...) {
+                                size_t source_length, const wchar_t *fmt, ...) {
     if (!errors) return true;
     parse_error_t error;
     error.source_start = source_location;
-    error.source_length = 0;
+    error.source_length = source_length;
     error.code = parse_error_syntax;
 
     va_list va;
@@ -856,6 +866,12 @@ static const wchar_t *error_format_for_character(wchar_t wc) {
         case VARIABLE_EXPAND_EMPTY: {
             return ERROR_NOT_PID;
         }
+        case BRACE_END:
+        case L'}':
+        case L',':
+        case BRACE_SEP: {
+            return ERROR_NO_VAR_NAME;
+        }
         default: {
             return ERROR_BAD_VAR_CHAR1;
         }
@@ -891,11 +907,11 @@ void parse_util_expand_variable_error(const wcstring &token, size_t global_token
             }
             if (looks_like_variable) {
                 append_syntax_error(
-                    errors, global_after_dollar_pos,
+                    errors, global_after_dollar_pos, 1,
                     double_quotes ? ERROR_BRACKETED_VARIABLE_QUOTED1 : ERROR_BRACKETED_VARIABLE1,
                     truncate(var_name, var_err_len).c_str());
             } else {
-                append_syntax_error(errors, global_after_dollar_pos, ERROR_BAD_VAR_CHAR1, L'{');
+                append_syntax_error(errors, global_after_dollar_pos, 1, ERROR_BAD_VAR_CHAR1, L'{');
             }
             break;
         }
@@ -903,11 +919,11 @@ void parse_util_expand_variable_error(const wcstring &token, size_t global_token
             // e.g.: echo foo"$"baz
             // These are only ever quotes, not command substitutions. Command substitutions are
             // handled earlier.
-            append_syntax_error(errors, global_dollar_pos, ERROR_NO_VAR_NAME);
+            append_syntax_error(errors, global_dollar_pos, 1, ERROR_NO_VAR_NAME);
             break;
         }
         case L'\0': {
-            append_syntax_error(errors, global_dollar_pos, ERROR_NO_VAR_NAME);
+            append_syntax_error(errors, global_dollar_pos, 1, ERROR_NO_VAR_NAME);
             break;
         }
         default: {
@@ -922,7 +938,7 @@ void parse_util_expand_variable_error(const wcstring &token, size_t global_token
             // arguments we pass but that's harmless.
             const wchar_t *error_fmt = error_format_for_character(token_stop_char);
 
-            append_syntax_error(errors, global_after_dollar_pos, error_fmt, token_stop_char);
+            append_syntax_error(errors, global_after_dollar_pos, 1, error_fmt, token_stop_char);
             break;
         }
     }
@@ -943,7 +959,53 @@ parser_test_error_bits_t parse_util_detect_errors_in_argument(const ast::argumen
     size_t source_start = source_range->start;
     parser_test_error_bits_t err = 0;
 
+    auto check_subtoken = [&arg_src, &out_errors, source_start](size_t begin, size_t end) -> int {
+        wcstring unesc;
+        if (!unescape_string(arg_src.c_str() + begin, end - begin, &unesc, UNESCAPE_SPECIAL)) {
+            if (out_errors) {
+                const wchar_t *fmt = L"Invalid token '%ls'";
+                if (arg_src.length() == 2 && arg_src[0] == L'\\' &&
+                        (arg_src[1] == L'c' || towlower(arg_src[1]) == L'u'
+                        || towlower(arg_src[1]) == L'x')) {
+                    fmt = L"Incomplete escape sequence '%ls'";
+                }
+
+                append_syntax_error(out_errors, source_start + begin, end - begin,
+                                    fmt, arg_src.c_str());
+            }
+            return 1;
+        }
+
+        parser_test_error_bits_t err = 0;
+        // Check for invalid variable expansions.
+        const size_t unesc_size = unesc.size();
+        for (size_t idx = 0; idx < unesc_size; idx++) {
+            if (unesc.at(idx) != VARIABLE_EXPAND && unesc.at(idx) != VARIABLE_EXPAND_SINGLE) {
+                continue;
+            }
+
+            wchar_t next_char = idx + 1 < unesc_size ? unesc.at(idx + 1) : L'\0';
+            if (next_char != VARIABLE_EXPAND && next_char != VARIABLE_EXPAND_SINGLE &&
+                next_char != '(' && !valid_var_name_char(next_char)) {
+                err = 1;
+                if (out_errors) {
+                    // We have something like $$$^....  Back up until we reach the first $.
+                    size_t first_dollar = idx;
+                    while (first_dollar > 0 &&
+                           (unesc.at(first_dollar - 1) == VARIABLE_EXPAND ||
+                            unesc.at(first_dollar - 1) == VARIABLE_EXPAND_SINGLE)) {
+                        first_dollar--;
+                    }
+                    parse_util_expand_variable_error(unesc, source_start, first_dollar, out_errors);
+                }
+            }
+        }
+
+        return err;
+    };
+
     size_t cursor = 0;
+    size_t checked = 0;
     wcstring subst;
 
     bool do_loop = true;
@@ -951,12 +1013,13 @@ parser_test_error_bits_t parse_util_detect_errors_in_argument(const ast::argumen
     while (do_loop) {
         size_t paren_begin = 0;
         size_t paren_end = 0;
+        bool has_dollar = false;
         switch (parse_util_locate_cmdsubst_range(arg_src, &cursor, &subst, &paren_begin, &paren_end,
-                                                 false, &is_quoted)) {
+                                                 false, &is_quoted, &has_dollar)) {
             case -1: {
                 err |= PARSER_TEST_ERROR;
                 if (out_errors) {
-                    append_syntax_error(out_errors, source_start, L"Mismatched parenthesis");
+                    append_syntax_error(out_errors, source_start, 1, L"Mismatched parenthesis");
                 }
                 return err;
             }
@@ -965,6 +1028,8 @@ parser_test_error_bits_t parse_util_detect_errors_in_argument(const ast::argumen
                 break;
             }
             case 1: {
+                err |= check_subtoken(checked, paren_begin - has_dollar);
+
                 assert(paren_begin < paren_end && "Parens out of order?");
                 parse_error_list_t subst_errors;
                 err |= parse_util_detect_errors(subst, &subst_errors);
@@ -978,6 +1043,8 @@ parser_test_error_bits_t parse_util_detect_errors_in_argument(const ast::argumen
                 if (out_errors != nullptr) {
                     out_errors->insert(out_errors->end(), subst_errors.begin(), subst_errors.end());
                 }
+
+                checked = paren_end + 1;
                 break;
             }
             default: {
@@ -985,37 +1052,7 @@ parser_test_error_bits_t parse_util_detect_errors_in_argument(const ast::argumen
             }
         }
     }
-
-    wcstring unesc;
-    if (!unescape_string(arg_src, &unesc, UNESCAPE_SPECIAL)) {
-        if (out_errors) {
-            append_syntax_error(out_errors, source_start, L"Invalid token '%ls'", arg_src.c_str());
-        }
-        return 1;
-    }
-
-    // Check for invalid variable expansions.
-    const size_t unesc_size = unesc.size();
-    for (size_t idx = 0; idx < unesc_size; idx++) {
-        if (unesc.at(idx) != VARIABLE_EXPAND && unesc.at(idx) != VARIABLE_EXPAND_SINGLE) {
-            continue;
-        }
-
-        wchar_t next_char = idx + 1 < unesc_size ? unesc.at(idx + 1) : L'\0';
-        if (next_char != VARIABLE_EXPAND && next_char != VARIABLE_EXPAND_SINGLE &&
-            next_char != '(' && !valid_var_name_char(next_char)) {
-            err = 1;
-            if (out_errors) {
-                // We have something like $$$^....  Back up until we reach the first $.
-                size_t first_dollar = idx;
-                while (first_dollar > 0 && (unesc.at(first_dollar - 1) == VARIABLE_EXPAND ||
-                                            unesc.at(first_dollar - 1) == VARIABLE_EXPAND_SINGLE)) {
-                    first_dollar--;
-                }
-                parse_util_expand_variable_error(unesc, source_start, first_dollar, out_errors);
-            }
-        }
-    }
+    err |= check_subtoken(checked, arg_src.size());
 
     return err;
 }
@@ -1037,10 +1074,10 @@ static bool detect_errors_in_backgrounded_job(const ast::job_t &job,
     if (!job_conj) return false;
 
     if (job_conj->parent->try_as<if_clause_t>()) {
-        errored = append_syntax_error(parse_errors, source_range->start,
+        errored = append_syntax_error(parse_errors, source_range->start, source_range->length,
                                       BACKGROUND_IN_CONDITIONAL_ERROR_MSG);
     } else if (job_conj->parent->try_as<while_header_t>()) {
-        errored = append_syntax_error(parse_errors, source_range->start,
+        errored = append_syntax_error(parse_errors, source_range->start, source_range->length,
                                       BACKGROUND_IN_CONDITIONAL_ERROR_MSG);
     } else if (const ast::job_list_t *jlist = job_conj->parent->try_as<ast::job_list_t>()) {
         // This isn't very complete, e.g. we don't catch 'foo & ; not and bar'.
@@ -1059,6 +1096,7 @@ static bool detect_errors_in_backgrounded_job(const ast::job_t &job,
                     "Unexpected decorator keyword");
                 const wchar_t *deco_name = (deco->kw == parse_keyword_t::kw_and ? L"and" : L"or");
                 errored = append_syntax_error(parse_errors, deco->source_range().start,
+                                              deco->source_range().length,
                                               BOOL_AFTER_BACKGROUND_ERROR_MSG, deco_name);
             }
         }
@@ -1075,6 +1113,7 @@ static bool detect_errors_in_decorated_statement(const wcstring &buff_src,
     using namespace ast;
     bool errored = false;
     auto source_start = dst.source_range().start;
+    auto source_length = dst.source_range().length;
     const statement_decoration_t decoration = dst.decoration();
 
     // Determine if the first argument is help.
@@ -1108,7 +1147,8 @@ static bool detect_errors_in_decorated_statement(const wcstring &buff_src,
     // Check that we don't try to pipe through exec.
     bool is_in_pipeline = (pipe_pos != pipeline_position_t::none);
     if (is_in_pipeline && decoration == statement_decoration_t::exec) {
-        errored = append_syntax_error(parse_errors, source_start, EXEC_ERR_MSG, L"exec");
+        errored = append_syntax_error(parse_errors, source_start, source_length,
+                                      INVALID_PIPELINE_CMD_ERR_MSG, L"exec");
     }
 
     // This is a somewhat stale check that 'and' and 'or' are not in pipelines, except at the
@@ -1119,8 +1159,14 @@ static bool detect_errors_in_decorated_statement(const wcstring &buff_src,
         // commands.
         const wcstring &command = dst.command.source(buff_src, storage);
         if (command == L"and" || command == L"or") {
-            errored =
-                append_syntax_error(parse_errors, source_start, EXEC_ERR_MSG, command.c_str());
+            errored = append_syntax_error(parse_errors, source_start, source_length,
+                                          INVALID_PIPELINE_CMD_ERR_MSG, command.c_str());
+        }
+
+        // Similarly for time (#8841).
+        if (command == L"time") {
+            errored = append_syntax_error(parse_errors, source_start, source_length,
+                                          TIME_IN_PIPELINE_ERR_MSG);
         }
     }
 
@@ -1130,7 +1176,7 @@ static bool detect_errors_in_decorated_statement(const wcstring &buff_src,
     const wcstring &com = dst.command.source(buff_src, storage);
     if (com == L"$status") {
         errored =
-            append_syntax_error(parse_errors, source_start,
+            append_syntax_error(parse_errors, source_start, source_length,
                                 _(L"$status is not valid as a command. See `help conditions`"));
     }
 
@@ -1148,8 +1194,8 @@ static bool detect_errors_in_decorated_statement(const wcstring &buff_src,
 
         // Check that pipes are sound.
         if (!errored && parser_is_pipe_forbidden(command) && is_in_pipeline) {
-            errored =
-                append_syntax_error(parse_errors, source_start, EXEC_ERR_MSG, command.c_str());
+            errored = append_syntax_error(parse_errors, source_start, source_length,
+                                          INVALID_PIPELINE_CMD_ERR_MSG, command.c_str());
         }
 
         // Check that we don't break or continue from outside a loop.
@@ -1178,7 +1224,7 @@ static bool detect_errors_in_decorated_statement(const wcstring &buff_src,
 
             if (!found_loop) {
                 errored = append_syntax_error(
-                    parse_errors, source_start,
+                    parse_errors, source_start, source_length,
                     (command == L"break" ? INVALID_BREAK_ERR_MSG : INVALID_CONTINUE_ERR_MSG));
             }
         }
@@ -1189,13 +1235,16 @@ static bool detect_errors_in_decorated_statement(const wcstring &buff_src,
             if (expand_one(command, expand_flag::skip_cmdsubst, operation_context_t::empty(),
                            parse_errors) &&
                 !builtin_exists(unexp_command)) {
-                errored = append_syntax_error(parse_errors, source_start, UNKNOWN_BUILTIN_ERR_MSG,
-                                              unexp_command.c_str());
+                errored = append_syntax_error(parse_errors, source_start, source_length,
+                                              UNKNOWN_BUILTIN_ERR_MSG, unexp_command.c_str());
             }
         }
 
         if (parse_errors) {
-            parse_error_offset_source_start(&new_errors, source_start);
+            // The expansion errors here go from the *command* onwards,
+            // so we need to offset them by the *command* offset,
+            // excluding the decoration.
+            parse_error_offset_source_start(&new_errors, dst.command.source_range().start);
             vec_append(*parse_errors, std::move(new_errors));
         }
     }
@@ -1207,7 +1256,8 @@ static bool detect_errors_in_decorated_statement(const wcstring &buff_src,
 static bool detect_errors_in_block_redirection_list(
     const ast::argument_or_redirection_list_t &args_or_redirs, parse_error_list_t *out_errors) {
     if (const auto *first_arg = get_first_arg(args_or_redirs)) {
-        return append_syntax_error(out_errors, first_arg->source_range().start, END_ARG_ERR_MSG);
+        return append_syntax_error(out_errors, first_arg->source_range().start,
+                                   first_arg->source_range().length, END_ARG_ERR_MSG);
     }
     return false;
 }
