@@ -5,20 +5,20 @@
 #include <fcntl.h>
 #include <paths.h>
 #include <signal.h>
-#include <stdio.h>
 #include <time.h>
+#include <unistd.h>
 
 #include <cstring>
-#include <memory>
 #ifdef HAVE_SPAWN_H
 #include <spawn.h>
 #endif
-#include <cwchar>
+#include <string>
+#include <vector>
 
 #include "common.h"
 #include "exec.h"
+#include "fds.h"
 #include "flog.h"
-#include "io.h"
 #include "iothread.h"
 #include "job_group.h"
 #include "postfork.h"
@@ -256,10 +256,10 @@ bool posix_spawner_t::check_fail(int err) {
 }
 
 posix_spawner_t::~posix_spawner_t() {
-    if (attr_) {
+    if (attr_.has_value()) {
         posix_spawnattr_destroy(this->attr());
     }
-    if (actions_) {
+    if (actions_.has_value()) {
         posix_spawn_file_actions_destroy(this->actions());
     }
 }
@@ -281,10 +281,13 @@ posix_spawner_t::posix_spawner_t(const job_t *j, const dup2_list_t &dup2s) {
     // desired_pgid tracks the pgroup for the process. If it is none, the pgroup is left unchanged.
     // If it is zero, create a new pgroup from the pid. If it is >0, join that pgroup.
     maybe_t<pid_t> desired_pgid = none();
-    if (auto pgid = j->group->get_pgid()) {
-        desired_pgid = *pgid;
-    } else if (j->processes.front()->leads_pgrp) {
-        desired_pgid = 0;
+    {
+        auto pgid = j->group->get_pgid();
+        if (pgid.has_value()) {
+            desired_pgid = *pgid;
+        } else if (j->processes.front()->leads_pgrp) {
+            desired_pgid = 0;
+        }
     }
 
     // Set the handling for job control signals back to the default.
@@ -336,7 +339,7 @@ maybe_t<pid_t> posix_spawner_t::spawn(const char *cmd, char *const argv[], char
     pid_t pid = -1;
     if (check_fail(posix_spawn(&pid, cmd, &*actions_, &*attr_, argv, envp))) {
         // The shebang wasn't introduced until UNIX Seventh Edition, so if
-        // the kernel won't run the binary we hand it off to the intpreter
+        // the kernel won't run the binary we hand it off to the interpreter
         // after performing a binary safety check, recommended by POSIX: a
         // line needs to exist before the first \0 with a lowercase letter
         if (error_ == ENOEXEC && is_thompson_shell_script(cmd)) {
@@ -543,6 +546,13 @@ void safe_report_exec_error(int err, const char *actual_cmd, const char *const *
                        actual_cmd);
             break;
         }
+#ifdef EBADARCH
+        case EBADARCH: {
+            FLOGF_SAFE(exec, "Failed to execute process '%s': Bad CPU type in executable.",
+                       actual_cmd);
+            break;
+        }
+#endif
         default: {
             char errnum_buff[64];
             format_long_safe(errnum_buff, err);
