@@ -4,21 +4,20 @@
 // keeping an internal representation of the current screen contents and trying to find the most
 // efficient way for transforming that to the desired screen content.
 //
-// IWYU pragma: no_include <cstddef>
 #include "config.h"
 
 #include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <termios.h>
-#include <time.h>
 #include <unistd.h>
 
+#include <cstdint>
 #include <cstring>
 #include <cwchar>
 
 #if HAVE_CURSES_H
-#include <curses.h>
+#include <curses.h>  // IWYU pragma: keep
 #elif HAVE_NCURSES_H
 #include <ncurses.h>
 #elif HAVE_NCURSES_CURSES_H
@@ -38,10 +37,12 @@
 #include "env.h"
 #include "fallback.h"  // IWYU pragma: keep
 #include "flog.h"
+#include "global_safety.h"
 #include "highlight.h"
 #include "output.h"
 #include "pager.h"
 #include "screen.h"
+#include "termsize.h"
 
 /// The number of characters to indent new blocks.
 #define INDENT_STEP 4u
@@ -189,6 +190,7 @@ static bool is_two_byte_escape_seq(const wchar_t *code, size_t *resulting_length
 
 /// Generic VT100 CSI-style sequence. <esc>, followed by zero or more ASCII characters NOT in
 /// the range [@,_], followed by one character in that range.
+/// This will also catch color sequences.
 static bool is_csi_style_escape_seq(const wchar_t *code, size_t *resulting_length) {
     if (code[1] != L'[') {
         return false;
@@ -214,37 +216,6 @@ static bool is_csi_style_escape_seq(const wchar_t *code, size_t *resulting_lengt
     return true;
 }
 
-/// Detect whether the escape sequence sets foreground/background color. Note that 24-bit color
-/// sequences are detected by `is_csi_style_escape_seq()` if they use the ANSI X3.64 pattern for
-/// such sequences. This function only handles those escape sequences for setting color that rely on
-/// the terminfo definition and which might use a different pattern.
-static bool is_color_escape_seq(const wchar_t *code, size_t *resulting_length) {
-    if (!cur_term) return false;
-
-    // Detect these terminfo color escapes with parameter value up to max_colors, all of which
-    // don't move the cursor.
-    const char *const esc[] = {
-        set_a_foreground,
-        set_a_background,
-        set_foreground,
-        set_background,
-    };
-
-    for (auto p : esc) {
-        if (!p) continue;
-
-        for (int k = 0; k < max_colors; k++) {
-            size_t esc_seq_len = try_sequence(tparm(const_cast<char *>(p), k), code);
-            if (esc_seq_len) {
-                *resulting_length = esc_seq_len;
-                return true;
-            }
-        }
-    }
-
-    return false;
-}
-
 /// Detect whether the escape sequence sets one of the terminal attributes that affects how text is
 /// displayed other than the color.
 static bool is_visual_escape_seq(const wchar_t *code, size_t *resulting_length) {
@@ -285,9 +256,6 @@ maybe_t<size_t> escape_code_length(const wchar_t *code) {
     if (!found) found = is_three_byte_escape_seq(code, &esc_seq_len);
     if (!found) found = is_csi_style_escape_seq(code, &esc_seq_len);
     if (!found) found = is_two_byte_escape_seq(code, &esc_seq_len);
-    // Colors are the hardest to match, so we try last.
-    // (also tparm is *slow*, we should try to find a better replacement)
-    if (!found) found = is_color_escape_seq(code, &esc_seq_len);
 
     return found ? maybe_t<size_t>{esc_seq_len} : none();
 }
@@ -300,7 +268,7 @@ size_t layout_cache_t::escape_code_length(const wchar_t *code) {
     if (esc_seq_len) return esc_seq_len;
 
     auto found = ::escape_code_length(code);
-    if (found) {
+    if (found.has_value()) {
         this->add_escape_code(wcstring(code, *found));
         esc_seq_len = *found;
     }
@@ -342,7 +310,7 @@ static size_t measure_run_from(const wchar_t *input, size_t start, size_t *out_e
                                layout_cache_t &cache) {
     size_t width = 0;
     size_t idx = start;
-    for (idx = start; !is_run_terminator(input[idx]); idx++) {
+    for (; !is_run_terminator(input[idx]); idx++) {
         if (input[idx] == L'\x1B') {
             // This is the start of an escape code; we assume it has width 0.
             // -1 because we are going to increment in the loop.
@@ -1291,7 +1259,7 @@ void screen_t::reset_abandoning_line(int screen_width) {
             std::string dim = fish_tparm(const_cast<char *>(enter_dim_mode));
             if (!dim.empty()) {
                 // Use dim if they have it, so the color will be based on their actual normal
-                // color and the background of the termianl.
+                // color and the background of the terminal.
                 abandon_line_string.append(str2wcstring(dim));
                 justgrey = false;
             }
