@@ -3,23 +3,19 @@
 
 #include "timer.h"
 
+#include <stdint.h>
+#include <stdio.h>
 #include <string.h>
 
 #include <algorithm>
-#include <cerrno>
 #include <chrono>
-#include <cstddef>
-#include <ctime>
+#include <cwchar>
+#include <functional>
+#include <string>
 
-#include "builtin.h"
 #include "common.h"
-#include "exec.h"
 #include "fallback.h"  // IWYU pragma: keep
-#include "io.h"
-#include "parser.h"
-#include "proc.h"
-#include "wgetopt.h"
-#include "wutil.h"  // IWYU pragma: keep
+#include "wutil.h"     // IWYU pragma: keep
 
 // Measuring time is always complicated with many caveats. Quite apart from the typical
 // gotchas faced by developers attempting to choose between monotonic vs non-monotonic and system vs
@@ -163,14 +159,14 @@ wcstring timer_snapshot_t::print_delta(const timer_snapshot_t &t1, const timer_s
 
     wcstring output;
     if (!verbose) {
-        append_format(output, L"\n_______________________________"
-                              L"\nExecuted in  %6.2F %s"
-                              L"\n   usr time  %6.2F %s"
-                              L"\n   sys time  %6.2F %s"
-                              L"\n",
-                              wall_time, unit_name(wall_unit),
-                              usr_time, unit_name(cpu_unit),
-                              sys_time, unit_name(cpu_unit));
+        append_format(output,
+                      L"\n_______________________________"
+                      L"\nExecuted in  %6.2F %s"
+                      L"\n   usr time  %6.2F %s"
+                      L"\n   sys time  %6.2F %s"
+                      L"\n",
+                      wall_time, unit_name(wall_unit), usr_time, unit_name(cpu_unit), sys_time,
+                      unit_name(cpu_unit));
     } else {
         auto fish_unit = get_unit(std::max(fish_sys_micros, fish_usr_micros));
         auto child_unit = get_unit(std::max(child_sys_micros, child_usr_micros));
@@ -179,8 +175,8 @@ wcstring timer_snapshot_t::print_delta(const timer_snapshot_t &t1, const timer_s
         double child_usr_time = convert(child_usr_micros, child_unit);
         double child_sys_time = convert(child_sys_micros, child_unit);
 
-        int column2_unit_len = std::max(strlen(unit_short_name(wall_unit)),
-                                        strlen(unit_short_name(cpu_unit)));
+        int column2_unit_len =
+            std::max(strlen(unit_short_name(wall_unit)), strlen(unit_short_name(cpu_unit)));
         append_format(output,
                       L"\n________________________________________________________"
                       L"\nExecuted in  %6.2F %-*s    %-*s  %s"
@@ -197,11 +193,7 @@ wcstring timer_snapshot_t::print_delta(const timer_snapshot_t &t1, const timer_s
     return output;
 };
 
-static std::vector<timer_snapshot_t> active_timers;
-
-static void pop_timer() {
-    auto t1 = active_timers.back();
-    active_timers.pop_back();
+static void timer_finished(const timer_snapshot_t &t1) {
     auto t2 = timer_snapshot_t::take();
 
     // Well, this is awkward. By defining `time` as a decorator and not a built-in, there's
@@ -212,6 +204,7 @@ static void pop_timer() {
 
 cleanup_t push_timer(bool enabled) {
     if (!enabled) return {[] {}};
-    active_timers.emplace_back(timer_snapshot_t::take());
-    return {[] { pop_timer(); }};
+
+    auto t1 = timer_snapshot_t::take();
+    return {[=] { timer_finished(t1); }};
 }
