@@ -7,22 +7,28 @@
 #include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <libgen.h>
 #include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <sys/mount.h>
 #include <sys/stat.h>
-#include <sys/statvfs.h>
 #include <sys/types.h>
 #include <unistd.h>
 #include <wctype.h>
 
-#include <atomic>
+#ifdef HAVE_XLOCALE_H
+#include <xlocale.h>
+#endif
+
+#include <algorithm>
 #include <cstring>
 #include <cwchar>
+#include <iterator>
+#include <locale>
+#include <mutex>
 #include <string>
 #include <unordered_map>
+#include <utility>
+#include <vector>
 
 #include "common.h"
 #include "fallback.h"  // IWYU pragma: keep
@@ -37,92 +43,6 @@ const file_id_t kInvalidFileID{};
 /// Map used as cache by wgettext.
 static owning_lock<std::unordered_map<wcstring, wcstring>> wgettext_map;
 
-bool wreaddir_resolving(DIR *dir, const wcstring &dir_path, wcstring &out_name, bool *out_is_dir) {
-    struct dirent *result = readdir(dir);
-    if (!result) {
-        out_name.clear();
-        return false;
-    }
-
-    out_name = str2wcstring(result->d_name);
-    if (!out_is_dir) {
-        return true;
-    }
-
-    // The caller cares if this is a directory, so check.
-    bool is_dir = false;
-    // We may be able to skip stat, if the readdir can tell us the file type directly.
-    bool check_with_stat = true;
-#ifdef HAVE_STRUCT_DIRENT_D_TYPE
-    if (result->d_type == DT_DIR) {
-        // Known directory.
-        is_dir = true;
-        check_with_stat = false;
-    } else if (result->d_type == DT_LNK || result->d_type == DT_UNKNOWN) {
-        // We want to treat symlinks to directories as directories. Use stat to resolve it.
-        check_with_stat = true;
-    } else {
-        // Regular file.
-        is_dir = false;
-        check_with_stat = false;
-    }
-#endif  // HAVE_STRUCT_DIRENT_D_TYPE
-    if (check_with_stat) {
-        // We couldn't determine the file type from the dirent; check by stat'ing it.
-        cstring fullpath = wcs2string(dir_path);
-        fullpath.push_back('/');
-        fullpath.append(result->d_name);
-        struct stat buf;
-        if (stat(fullpath.c_str(), &buf) != 0) {
-            is_dir = false;
-        } else {
-            is_dir = S_ISDIR(buf.st_mode);
-        }
-    }
-    *out_is_dir = is_dir;
-    return true;
-}
-
-bool wreaddir(DIR *dir, wcstring &out_name) {
-    struct dirent *result = readdir(dir);
-    if (!result) {
-        out_name.clear();
-        return false;
-    }
-
-    out_name = str2wcstring(result->d_name);
-    return true;
-}
-
-bool readdir_for_dirs(DIR *dir, std::string *out_name) {
-    struct dirent *result = nullptr;
-    while (!result) {
-        result = readdir(dir);
-        if (!result) break;
-
-#if HAVE_STRUCT_DIRENT_D_TYPE
-        switch (result->d_type) {
-            case DT_DIR:
-            case DT_LNK:
-            case DT_UNKNOWN: {
-                break;  // these may be directories
-            }
-            default: {
-                break;  // nothing else can
-            }
-        }
-#else
-        // We can't determine if it's a directory or not, so just return it.
-        break;
-#endif
-    }
-
-    if (result && out_name) {
-        *out_name = result->d_name;
-    }
-    return result != nullptr;
-}
-
 wcstring wgetcwd() {
     char cwd[PATH_MAX];
     char *res = getcwd(cwd, sizeof(cwd));
@@ -139,21 +59,174 @@ DIR *wopendir(const wcstring &name) {
     return opendir(tmp.c_str());
 }
 
-dir_t::dir_t(const wcstring &path) {
-    const cstring tmp = wcs2string(path);
-    this->dir = opendir(tmp.c_str());
+#ifdef HAVE_STRUCT_DIRENT_D_TYPE
+static maybe_t<dir_entry_type_t> dirent_type_to_entry_type(uint8_t dt) {
+    switch (dt) {
+        case DT_FIFO:
+            return dir_entry_type_t::fifo;
+        case DT_CHR:
+            return dir_entry_type_t::chr;
+        case DT_DIR:
+            return dir_entry_type_t::dir;
+        case DT_BLK:
+            return dir_entry_type_t::blk;
+        case DT_REG:
+            return dir_entry_type_t::reg;
+        case DT_LNK:
+            return dir_entry_type_t::lnk;
+        case DT_SOCK:
+            return dir_entry_type_t::sock;
+#if defined(DT_WHT)
+        // OpenBSD doesn't have this one
+        case DT_WHT:
+            return dir_entry_type_t::whiteout;
+#endif
+        case DT_UNKNOWN:
+        default:
+            return none();
+    }
 }
+#endif
 
-dir_t::~dir_t() {
-    if (this->dir != nullptr) {
-        closedir(this->dir);
-        this->dir = nullptr;
+static maybe_t<dir_entry_type_t> stat_mode_to_entry_type(mode_t m) {
+    switch (m & S_IFMT) {
+        case S_IFIFO:
+            return dir_entry_type_t::fifo;
+        case S_IFCHR:
+            return dir_entry_type_t::chr;
+        case S_IFDIR:
+            return dir_entry_type_t::dir;
+        case S_IFBLK:
+            return dir_entry_type_t::blk;
+        case S_IFREG:
+            return dir_entry_type_t::reg;
+        case S_IFLNK:
+            return dir_entry_type_t::lnk;
+        case S_IFSOCK:
+            return dir_entry_type_t::sock;
+#if defined(S_IFWHT)
+        case S_IFWHT:
+            return dir_entry_type_t::whiteout;
+#endif
+        default:
+            return none();
     }
 }
 
-bool dir_t::valid() const { return this->dir != nullptr; }
+dir_iter_t::entry_t::entry_t() = default;
+dir_iter_t::entry_t::~entry_t() = default;
 
-bool dir_t::read(wcstring &name) const { return wreaddir(this->dir, name); }
+void dir_iter_t::entry_t::reset() {
+    this->name.clear();
+    this->inode = {};
+    this->type_.reset();
+    this->stat_.reset();
+}
+
+maybe_t<dir_entry_type_t> dir_iter_t::entry_t::check_type() const {
+    // Call stat if needed to populate our type, swallowing errors.
+    if (!this->type_) {
+        this->do_stat();
+    }
+    return this->type_;
+}
+
+const maybe_t<struct stat> &dir_iter_t::entry_t::stat() const {
+    if (!stat_) {
+        (void)this->do_stat();
+    }
+    return stat_;
+}
+
+void dir_iter_t::entry_t::do_stat() const {
+    // We want to set both our type and our stat buffer.
+    // If we follow symlinks and stat() errors with a bad symlink, set the type to link, but do not
+    // populate the stat buffer.
+    if (this->dirfd_ < 0) {
+        return;
+    }
+    std::string narrow = wcs2string(this->name);
+    struct stat s {};
+    if (fstatat(this->dirfd_, narrow.c_str(), &s, 0) == 0) {
+        this->stat_ = s;
+        this->type_ = stat_mode_to_entry_type(s.st_mode);
+    } else {
+        switch (errno) {
+            case ELOOP:
+                this->type_ = dir_entry_type_t::lnk;
+                break;
+
+            case EACCES:
+            case EIO:
+            case ENOENT:
+            case ENOTDIR:
+            case ENAMETOOLONG:
+                // These are "expected" errors.
+                this->type_ = none();
+                break;
+
+            default:
+                wperror(L"fstatat");
+                break;
+        }
+    }
+}
+
+dir_iter_t::dir_iter_t(const wcstring &path) {
+    dir_.reset(wopendir(path));
+    if (!dir_) {
+        error_ = errno;
+        return;
+    }
+    entry_.dirfd_ = dirfd(&*dir_);
+}
+
+dir_iter_t::dir_iter_t(dir_iter_t &&rhs) { *this = std::move(rhs); }
+
+dir_iter_t &dir_iter_t::operator=(dir_iter_t &&rhs) {
+    // Steal the fields; ensure rhs no longer has DIR* and forgets its fd.
+    this->dir_ = std::move(rhs.dir_);
+    this->error_ = rhs.error_;
+    this->entry_ = std::move(rhs.entry_);
+    rhs.dir_ = nullptr;
+    rhs.entry_.dirfd_ = -1;
+    return *this;
+}
+
+void dir_iter_t::rewind() {
+    if (dir_) {
+        rewinddir(&*dir_);
+    }
+}
+
+dir_iter_t::~dir_iter_t() = default;
+
+const dir_iter_t::entry_t *dir_iter_t::next() {
+    if (!dir_) {
+        return nullptr;
+    }
+    errno = 0;
+    struct dirent *dent = readdir(&*dir_);
+    if (!dent) {
+        error_ = errno;
+        return nullptr;
+    }
+    // Skip . and ..
+    if (!strcmp(dent->d_name, ".") || !strcmp(dent->d_name, "..")) {
+        return next();
+    }
+    entry_.reset();
+    entry_.name = str2wcstring(dent->d_name);
+    entry_.inode = dent->d_ino;
+#ifdef HAVE_STRUCT_DIRENT_D_TYPE
+    auto type = dirent_type_to_entry_type(dent->d_type);
+    // Do not store symlinks as we will need to resolve them.
+    if (type != dir_entry_type_t::lnk) {
+        entry_.type_ = type;
+    }
+#endif
+    return &entry_;
+}
 
 int wstat(const wcstring &file_name, struct stat *buf) {
     const cstring tmp = wcs2string(file_name);
@@ -500,6 +573,15 @@ ssize_t wwrite_to_fd(const wchar_t *input, size_t input_len, int fd) {
     return success ? total_written : -1;
 }
 
+enum : wint_t {
+    PUA1_START = 0xE000,
+    PUA1_END = 0xF900,
+    PUA2_START = 0xF0000,
+    PUA2_END = 0xFFFFE,
+    PUA3_START = 0x100000,
+    PUA3_END = 0x10FFFE,
+};
+
 /// Return one if the code point is in a Unicode private use area.
 static int fish_is_pua(wint_t wc) {
     if (PUA1_START <= wc && wc < PUA1_END) return 1;
@@ -690,11 +772,10 @@ unsigned long long fish_wcstoull(const wchar_t *str, const wchar_t **endptr, int
 
 /// Like wcstod(), but wcstod() is enormously expensive on some platforms so this tries to have a
 /// fast path.
-double fish_wcstod(const wchar_t *str, wchar_t **endptr) {
+double fish_wcstod(const wchar_t *str, wchar_t **endptr, size_t len) {
     // We can ignore the locale because we use LC_NUMERIC=C!
     // The "fast path." If we're all ASCII and we fit inline, use strtod().
     char narrow[128];
-    size_t len = std::wcslen(str);
     size_t len_plus_0 = 1 + len;
     auto is_ascii = [](wchar_t c) { return 0 <= c && c <= 127; };
     if (len_plus_0 <= sizeof narrow && std::all_of(str, str + len, is_ascii)) {
@@ -711,6 +792,75 @@ double fish_wcstod(const wchar_t *str, wchar_t **endptr) {
     return std::wcstod(str, endptr);
 }
 
+double fish_wcstod(const wchar_t *str, wchar_t **endptr) {
+    return fish_wcstod(str, endptr, std::wcslen(str));
+}
+double fish_wcstod(const wcstring &str, wchar_t **endptr) {
+    return fish_wcstod(str.c_str(), endptr, str.size());
+}
+
+/// Like wcstod(), but allows underscore separators. Leading, trailing, and multiple underscores are
+/// allowed, as are underscores next to decimal (.), exponent (E/e/P/p), and hexadecimal (X/x)
+/// delimiters. This consumes trailing underscores -- endptr will point past the last underscore
+/// which is legal to include in a parse (according to the above rules). Free-floating leading
+/// underscores ("_ 3") are not allowed and will result in a no-parse. Underscores are not allowed
+/// before or inside of "infinity" or "nan" input. Trailing underscores after "infinity" or "nan"
+/// are not consumed.
+double fish_wcstod_underscores(const wchar_t *str, wchar_t **endptr) {
+    const wchar_t *orig = str;
+    while (iswspace(*str)) str++;  // Skip leading whitespace.
+    size_t leading_whitespace = size_t(str - orig);
+    auto is_sign = [](wchar_t c) { return c == L'+' || c == L'-'; };
+    auto is_inf_or_nan_char = [](wchar_t c) {
+        return c == L'i' || c == L'I' || c == L'n' || c == L'N';
+    };
+    // We don't do any underscore-stripping for infinity/NaN.
+    if (is_inf_or_nan_char(*str) || (is_sign(*str) && is_inf_or_nan_char(*(str + 1)))) {
+        return fish_wcstod(orig, endptr);
+    }
+    // We build a string to pass to the system wcstod, pruned of underscores. We will take all
+    // leading alphanumeric characters that can appear in a strtod numeric literal, dots (.), and
+    // signs (+/-). In order to be more clever, for example to stop earlier in the case of strings
+    // like "123xxxxx", we would need to do a full parse, because sometimes 'a' is a hex digit and
+    // sometimes it is the end of the parse, sometimes a dot '.' is a decimal delimiter and
+    // sometimes it is the end of the valid parse, as in "1_2.3_4.5_6", etc.
+    wcstring pruned;
+    // We keep track of the positions *in the pruned string* where there used to be underscores. We
+    // will pass the pruned version of the input string to the system wcstod, which in turn will
+    // tell us how many characters it consumed. Then we will set our own endptr based on (1) the
+    // number of characters consumed from the pruned string, and (2) how many underscores came
+    // before the last consumed character. The alternative to doing it this way (for example, "only
+    // deleting the correct underscores") would require actually parsing the input string, so that
+    // we can know when to stop grabbing characters and dropping underscores, as in "1_2.3_4.5_6".
+    std::vector<size_t> underscores;
+    // If we wanted to future-proof against a strtod from the future that, say, allows octal
+    // literals using 0o, etc., we could just use iswalnum, instead of iswxdigit and P/p/X/x checks.
+    while (iswxdigit(*str) || *str == L'P' || *str == L'p' || *str == L'X' || *str == L'x' ||
+           is_sign(*str) || *str == L'.' || *str == L'_') {
+        if (*str == L'_') {
+            underscores.push_back(pruned.length());
+        } else {
+            pruned.push_back(*str);
+        }
+        str++;
+    }
+    const wchar_t *pruned_begin = pruned.c_str();
+    const wchar_t *pruned_end = nullptr;
+    double result = fish_wcstod(pruned_begin, (wchar_t **)(&pruned_end));
+    if (pruned_end == pruned_begin) {
+        if (endptr) *endptr = (wchar_t *)orig;
+        return result;
+    }
+    auto consumed_underscores_end =
+        std::upper_bound(underscores.begin(), underscores.end(), size_t(pruned_end - pruned_begin));
+    size_t num_underscores_consumed = std::distance(underscores.begin(), consumed_underscores_end);
+    if (endptr) {
+        *endptr = (wchar_t *)(orig + leading_whitespace + (pruned_end - pruned_begin) +
+                              num_underscores_consumed);
+    }
+    return result;
+}
+
 file_id_t file_id_t::from_stat(const struct stat &buf) {
     file_id_t result = {};
     result.device = buf.st_dev;
@@ -792,6 +942,22 @@ static int compare(T a, T b) {
     return 0;
 }
 
+/// \return true if \param rhs has higher mtime seconds than this file_id_t.
+/// If identical, nanoseconds are compared.
+bool file_id_t::older_than(const file_id_t &rhs) const {
+    int ret = compare(mod_seconds, rhs.mod_seconds);
+    if (!ret) ret = compare(mod_nanoseconds, rhs.mod_nanoseconds);
+    switch (ret) {
+        case -1:
+            return true;
+        case 1:
+        case 0:
+            return false;
+        default:
+            DIE("unreachable");
+    }
+}
+
 int file_id_t::compare_file_id(const file_id_t &rhs) const {
     // Compare each field, stopping when we get to a non-equal field.
     int ret = 0;
